from datetime import datetime, timedelta, timezone
from typing import Literal

from lite_dist2.expections import LD2ModelTypeError
from lite_dist2.type_definitions import PortableValueType, PrimitiveValueType


def hex2int(hex_str: str) -> int:
    return int(hex_str, base=16)


def int2hex(val: int) -> str:
    return hex(val)


def hex2float(hex_str: str) -> float:
    return float.fromhex(hex_str)


def float2hex(val: float) -> str:
    return val.hex()


def numerize(type_name: Literal["bool", "int", "float"], value: PortableValueType) -> PrimitiveValueType:
    match type_name:
        case "bool":
            return value
        case "int":
            return hex2int(value)
        case "float":
            return hex2float(value)
        case _:
            raise LD2ModelTypeError(type_name)


def portablize(type_name: Literal["bool", "int", "float"], value: PrimitiveValueType) -> PortableValueType:
    match type_name:
        case "bool":
            return value
        case "int":
            return int2hex(value)
        case "float":
            return float2hex(value)
        case _:
            raise LD2ModelTypeError(type_name)


def publish_timestamp() -> datetime:
    return datetime.now(tz=timezone(timedelta(hours=+9), "JST"))



from __future__ import annotations

from pydantic import BaseModel, Field

from lite_dist2.curriculum_models.study_portables import StudyStorage, StudySummary
from lite_dist2.curriculum_models.study_status import StudyStatus
from lite_dist2.curriculum_models.trial import TrialModel


class BaseTableResponse(BaseModel):
    pass


class OkResponse(BaseTableResponse):
    ok: bool = Field(...)


class TrialReserveResponse(BaseTableResponse):
    trial: TrialModel | None = Field(
        ...,
        description=(
            "Reserved trial for the worker node. "
            "None if the curriculum is empty or no trial which can be processed by the worker node's capabilities."
        ),
    )


class StudyRegisteredResponse(BaseTableResponse):
    study_id: str = Field(
        ...,
        description="Published `study_id` of registered study.",
    )


class StudyResponse(BaseTableResponse):
    status: StudyStatus = Field(
        ...,
        description="Status of the target Study.",
    )
    result: StudyStorage | None = Field(
        None,
        description="Results of completed study. If the study is not completed or not found, then `None`.",
    )


class CurriculumSummaryResponse(BaseTableResponse):
    summaries: list[StudySummary] = Field(
        ...,
        description="The list of study (containing storage) summary.",
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from lite_dist2.study_strategies import BaseStudyStrategy, StudyStrategyModel

if TYPE_CHECKING:
    from lite_dist2.curriculum_models.trial import Mapping
    from lite_dist2.curriculum_models.trial_table import TrialTable
    from lite_dist2.study_strategies.base_study_strategy import StudyStrategyParam
    from lite_dist2.value_models.aligned_space import ParameterAlignedSpace


class FindExactStudyStrategy(BaseStudyStrategy):
    def __init__(self, study_strategy_param: StudyStrategyParam | None) -> None:
        super().__init__(study_strategy_param)
        self.found_mapping: Mapping | None = None

    def is_done(self, trial_table: TrialTable, _parameter_space: ParameterAlignedSpace) -> bool:
        if self.found_mapping:
            return True
        self.found_mapping = self._find(trial_table)
        return bool(self.found_mapping)

    def _find(self, trial_table: TrialTable) -> Mapping | None:
        return trial_table.find_target_value(self.study_strategy_param.target_value)

    def extract_mappings(self, trial_table: TrialTable) -> list[Mapping]:
        if not self.found_mapping:
            self.found_mapping = self._find(trial_table)
        if not self.found_mapping:
            return []
        return [self.found_mapping]

    def to_model(self) -> StudyStrategyModel:
        return StudyStrategyModel(
            type="find_exact",
            study_strategy_param=self.study_strategy_param,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from lite_dist2.expections import LD2InvalidSpaceError, LD2ParameterError
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace, ParameterAlignedSpaceModel
from lite_dist2.value_models.base_space import FlattenSegment
from lite_dist2.value_models.line_segment import LineSegment, LineSegmentModel, ParameterRangeBool, ParameterRangeInt

if TYPE_CHECKING:
    from lite_dist2.type_definitions import PrimitiveValueType


@pytest.mark.parametrize(
    ("axes", "message"),
    [
        (
            [
                ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
                ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=200, start=0),
            ],
            "Invalid space: Filling from lower dimension",
        ),
        (
            [
                ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=200, start=0),
                ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=200, start=0),
            ],
            "Invalid space: Upper dimension must be size=1",
        ),
        (
            [
                ParameterRangeInt(name="z", type="int", size=100, ambient_index=0, ambient_size=200, start=0),
                ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=200, start=0),
                ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
            ],
            "Invalid space: Upper dimension must be size=1",
        ),
        (
            [
                ParameterRangeInt(name="z", type="int", size=1, ambient_index=0, ambient_size=200, start=0),
                ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=200, start=0),
                ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=200, start=0),
            ],
            "Invalid space: Upper dimension must be size=1",
        ),
        (
            [
                ParameterRangeInt(name="z", type="int", size=1, ambient_index=0, ambient_size=200, start=0),
                ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=None, start=0),
                ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=200, start=0),
            ],
            "Invalid space: Infinite dimension is only allowed at the first dimension",
        ),
    ],
)
def test_parameter_space_fill_from_lower_raise(
    axes: list[LineSegment],
    message: str,
) -> None:
    with pytest.raises(LD2InvalidSpaceError, match=message):
        _ = ParameterAlignedSpace(axes, check_lower_filling=True)


@pytest.mark.parametrize(
    "axes",
    [
        [
            ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=2, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="y", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="y", type="int", size=2, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=2, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="y", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="z", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="y", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="z", type="int", size=2, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="y", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="z", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="y", type="int", size=2, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="z", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=None, start=0),
        ],
        [
            ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=None, start=0),
            ParameterRangeInt(name="y", type="int", size=2, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="z", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
        ],
    ],
)
def test_parameter_space_fill_from_lower_not_raise(axes: list[LineSegment]) -> None:
    space = ParameterAlignedSpace(axes=axes, check_lower_filling=True)
    assert space is not None


def test_parameter_space_get_flatten_ambient_start_and_size_raise() -> None:
    space = ParameterAlignedSpace(
        axes=[ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0)],
        check_lower_filling=False,
    )
    with pytest.raises(LD2InvalidSpaceError, match="Invalid space: Cannot get flatten info."):
        _ = space.get_flatten_ambient_start_and_size()


@pytest.mark.parametrize(
    ("space", "expected"),
    [
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
                ],
                check_lower_filling=True,
            ),
            FlattenSegment(0, 1),
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=2, ambient_index=10, ambient_size=100, start=0),
                ],
                check_lower_filling=True,
            ),
            FlattenSegment(10, 2),
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
                    ParameterRangeInt(name="y", type="int", size=13, ambient_index=0, ambient_size=100, start=0),
                ],
                check_lower_filling=True,
            ),
            FlattenSegment(0, 13),
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
                    ParameterRangeInt(name="y", type="int", size=13, ambient_index=0, ambient_size=100, start=0),
                    ParameterRangeInt(name="z", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
                ],
                check_lower_filling=True,
            ),
            FlattenSegment(0, 1300),
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
                    ParameterRangeInt(name="y", type="int", size=1, ambient_index=17, ambient_size=100, start=0),
                    ParameterRangeInt(name="z", type="int", size=19, ambient_index=81, ambient_size=100, start=0),
                ],
                check_lower_filling=True,
            ),
            FlattenSegment(1781, 19),
        ),
    ],
)
def test_parameter_space_get_flatten_ambient_start_and_size(
    space: ParameterAlignedSpace,
    expected: FlattenSegment,
) -> None:
    actual = space.get_flatten_ambient_start_and_size()
    assert actual == expected


@pytest.mark.parametrize(
    ("space", "expected"),
    [
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
                ],
                check_lower_filling=True,
            ),
            (1,),
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeBool(type="bool", size=2, ambient_index=0, ambient_size=2, start=False),
                    ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
                    ParameterRangeInt(name="y", type="int", size=70, ambient_index=0, ambient_size=70, start=0),
                ],
                check_lower_filling=True,
            ),
            (7000, 70, 1),
        ),
    ],
)
def test_parameter_aligned_space_model_lower_element_num_by_dim(
    space: ParameterAlignedSpace,
    expected: tuple[int, ...],
) -> None:
    actual = space.lower_element_num_by_dim()
    assert actual == expected


@pytest.mark.parametrize(
    ("space", "start_and_sizes", "expected"),
    [
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeBool(type="bool", size=2, ambient_index=0, ambient_size=2, start=False),
                    ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
                    ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
                ],
                check_lower_filling=False,
            ),
            [(0, 1), (10, 10), (0, 100)],
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeBool(type="bool", size=1, ambient_index=0, ambient_size=2, start=False),
                    ParameterRangeInt(name="x", type="int", size=10, ambient_index=10, ambient_size=100, start=10),
                    ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
                ],
                check_lower_filling=False,
            ),
        ),
    ],
)
def test_parameter_aligned_space_slice(
    space: ParameterAlignedSpace,
    start_and_sizes: list[tuple[int, int]],
    expected: ParameterAlignedSpace,
) -> None:
    actual = space.slice(start_and_sizes)
    assert actual == expected


def test_parameter_aligned_space_slice_raise_inconsistent_start_and_sizes() -> None:
    space = ParameterAlignedSpace(
        axes=[
            ParameterRangeBool(type="bool", size=2, ambient_index=0, ambient_size=2, start=False),
            ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
            ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, ambient_size=100, start=0),
        ],
        check_lower_filling=False,
    )
    ss = [(0, 1)]

    with pytest.raises(LD2ParameterError):
        _ = space.slice(ss)


@pytest.mark.parametrize(
    ("one", "other", "target_dim", "expected"),
    [
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=100, ambient_index=102, start=101, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            0,
            False,  # not derived by same ambient space False
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, start=0, ambient_size=100),
                    ParameterRangeInt(name="y", type="int", size=100, ambient_index=100, start=100, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            0,
            False,  # not same filling dim False
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, start=0, ambient_size=100),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, start=0, ambient_size=100),
                ],
                check_lower_filling=False,
            ),
            1,
            False,  # try merge filling dim
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ParameterRangeInt(name="y", type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(name="x", type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ParameterRangeInt(name="y", type="int", size=100, ambient_index=100, start=100, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            0,
            False,  # try merge shallower dim than not filled dim
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=100, start=100, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            0,
            True,  # adjacency True
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=100, start=100, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            0,
            True,  # adjacency reverse True
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=101, start=101, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            0,
            False,  # has stride False
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=101, start=101, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            0,
            False,  # has stride reverse False
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=100, start=100, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
            0,
            True,  # filled y True
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=100, start=100, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
            0,
            True,  # filled y reverse True
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=101, start=101, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
            0,
            False,  # filled y but stride x False
        ),
        (
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=101, start=101, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
            0,
            False,  # filled y but stride x reverse False
        ),
    ],
)
def test_parameter_aligned_space_can_merge(
    one: ParameterAlignedSpace,
    other: ParameterAlignedSpace,
    target_dim: int,
    expected: bool,
) -> None:
    actual = one.can_merge(other, target_dim)
    actual_reversed = other.can_merge(one, target_dim)
    assert actual == expected
    assert actual_reversed == expected


@pytest.mark.parametrize(
    ("one", "other", "target_dim", "expected"),
    [
        (  # adjacency
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=100, start=100, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
            0,
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=200, ambient_index=0, start=0, ambient_size=None),
                ],
                check_lower_filling=False,
            ),
        ),
        (  # filled y
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=100, ambient_index=100, start=100, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
            0,
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=200, ambient_index=0, start=0, name="x", ambient_size=None),
                    ParameterRangeInt(type="int", size=1000, ambient_index=0, start=0, name="y", ambient_size=1000),
                ],
                check_lower_filling=False,
            ),
        ),
    ],
)
def test_parameter_aligned_space_merge(
    one: ParameterAlignedSpace,
    other: ParameterAlignedSpace,
    target_dim: int,
    expected: ParameterAlignedSpace,
) -> None:
    actual = one.merge(other, target_dim)
    actual_reversed = other.merge(one, target_dim)
    assert actual == expected
    assert actual_reversed == expected


@pytest.mark.parametrize(
    ("flatten_index", "lower_element_num_by_dim", "expected"),
    [
        (0, (1,), (0,)),
        (0, (24, 1), (0, 0)),
        (591, (100, 10, 1), (5, 9, 1)),
        (591, (45, 9, 1), (13, 0, 6)),
    ],
)
def test_parameter_aligned_space_model_loom_by_flatten_index(
    flatten_index: int,
    lower_element_num_by_dim: tuple[int, ...],
    expected: tuple[int, ...],
) -> None:
    actual = ParameterAlignedSpace.loom_by_flatten_index(flatten_index, lower_element_num_by_dim)
    assert actual == expected


@pytest.mark.parametrize(
    ("space", "expected"),
    [
        pytest.param(
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=5, ambient_index=0, ambient_size=5, start=0),
                ],
                check_lower_filling=True,
            ),
            [(0,), (1,), (2,), (3,), (4,)],
            id="1D",
        ),
        pytest.param(
            ParameterAlignedSpace(
                axes=[
                    ParameterRangeInt(type="int", size=5, ambient_index=0, ambient_size=5, start=0),
                    ParameterRangeInt(type="int", size=4, ambient_index=0, ambient_size=4, start=0),
                ],
                check_lower_filling=True,
            ),
            [
                (0, 0),
                (0, 1),
                (0, 2),
                (0, 3),
                (1, 0),
                (1, 1),
                (1, 2),
                (1, 3),
                (2, 0),
                (2, 1),
                (2, 2),
                (2, 3),
                (3, 0),
                (3, 1),
                (3, 2),
                (3, 3),
                (4, 0),
                (4, 1),
                (4, 2),
                (4, 3),
            ],
            id="2D",
        ),
    ],
)
def test_parameter_aligned_space_model_grid(
    space: ParameterAlignedSpace,
    expected: list[tuple[PrimitiveValueType, ...]],
) -> None:
    actual = list(space.grid())
    assert actual == expected


@pytest.mark.parametrize(
    "model",
    [
        ParameterAlignedSpaceModel(
            type="aligned",
            axes=[
                LineSegmentModel(
                    name="x",
                    type="float",
                    size="0xa",
                    step="0x1.0000000000000p+1",
                    start="-0x1.4000000000000p+2",
                    ambient_size="0x14",
                    ambient_index="0x5",
                ),
                LineSegmentModel(
                    name="y",
                    type="int",
                    size="0xa",
                    step="0x2",
                    start="-0x5",
                    ambient_size="0x14",
                    ambient_index="0x5",
                ),
                LineSegmentModel(
                    name="z",
                    type="bool",
                    size="0x2",
                    step="0x1",
                    start=False,
                    ambient_index="0x0",
                    ambient_size="0x2",
                ),
            ],
            check_lower_filling=False,
        ),
        ParameterAlignedSpaceModel(
            type="aligned",
            axes=[
                LineSegmentModel(
                    name="x",
                    type="float",
                    size="0x1",
                    step="0x1.0000000000000p+1",
                    start="-0x1.4000000000000p+2",
                    ambient_size="0x14",
                    ambient_index="0x5",
                ),
                LineSegmentModel(
                    name="y",
                    type="int",
                    size="0x1",
                    step="0x2",
                    start="-0x5",
                    ambient_size="0x14",
                    ambient_index="0x5",
                ),
                LineSegmentModel(
                    name="z",
                    type="bool",
                    size="0x2",
                    step="0x1",
                    start=False,
                    ambient_index="0x0",
                    ambient_size="0x2",
                ),
            ],
            check_lower_filling=True,
        ),
    ],
)
def test_parameter_aligned_space_model_to_model_from_model(model: ParameterAlignedSpaceModel) -> None:
    space = ParameterAlignedSpace.from_model(model)
    reconstructed_model = space.to_model()
    assert model == reconstructed_model



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

if TYPE_CHECKING:
    from lite_dist2.curriculum_models.trial_table import TrialTable
    from lite_dist2.value_models.aligned_space import ParameterAlignedSpace
    from lite_dist2.value_models.base_space import ParameterSpace


class SuggestStrategyParam(BaseModel):
    strict_aligned: bool


class SuggestStrategyModel(BaseModel):
    type: Literal["sequential", "random", "designated"]
    suggest_strategy_param: SuggestStrategyParam


class BaseSuggestStrategy(metaclass=abc.ABCMeta):
    def __init__(
        self,
        suggest_parameter: SuggestStrategyParam,
        parameter_space: ParameterAlignedSpace,
    ) -> None:
        self.suggest_parameter = suggest_parameter
        self.parameter_space = parameter_space

    @abc.abstractmethod
    def to_model(self) -> SuggestStrategyModel:
        pass

    @abc.abstractmethod
    def suggest(self, trial_table: TrialTable, max_num: int) -> ParameterSpace:
        pass



import itertools
from typing import Any

import pytest

from lite_dist2.value_models.aligned_space import ParameterAlignedSpace
from lite_dist2.value_models.base_space import FlattenSegment
from lite_dist2.value_models.line_segment import ParameterRangeFloat, ParameterRangeInt
from lite_dist2.value_models.parameter_aligned_space_helper import infinite_product, remap_space, simplify


@pytest.mark.parametrize(
    ("sub_spaces", "target_dim", "expected"),
    [
        pytest.param(
            [],
            0,
            [],
            id="empty",
        ),
        pytest.param(
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            0,
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            id="single",
        ),
        pytest.param(
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=200, start=200, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            0,
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=200, start=200, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            id="far double",
        ),
        pytest.param(
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=200, start=200, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            0,
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=200, start=200, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            id="far double reversed",
        ),
        pytest.param(
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=100, start=100, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            0,
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=200, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            id="adjacency double",
        ),
        pytest.param(
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=100, start=100, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            0,
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=200, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            id="adjacency reversed double",
        ),
        pytest.param(
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=100, start=100, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=100, ambient_index=200, start=200, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            0,
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=300, ambient_index=0, start=0, ambient_size=None),
                    ],
                    check_lower_filling=False,
                ),
            ],
            id="adjacency triple",
        ),
        pytest.param(
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
                        ParameterRangeInt(name="y", type="int", size=10, ambient_index=0, ambient_size=20, start=0),
                    ],
                    check_lower_filling=True,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=1, ambient_index=1, ambient_size=100, start=0),
                        ParameterRangeInt(name="y", type="int", size=10, ambient_index=10, ambient_size=20, start=10),
                    ],
                    check_lower_filling=True,
                ),
            ],
            1,
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
                        ParameterRangeInt(name="y", type="int", size=10, ambient_index=0, ambient_size=20, start=0),
                    ],
                    check_lower_filling=True,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=1, ambient_index=1, ambient_size=100, start=0),
                        ParameterRangeInt(name="y", type="int", size=10, ambient_index=10, ambient_size=20, start=10),
                    ],
                    check_lower_filling=True,
                ),
            ],
            id="false adjacency (different dimension)",
        ),
        pytest.param(
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
                        ParameterRangeInt(name="y", type="int", size=10, ambient_index=0, ambient_size=20, start=0),
                    ],
                    check_lower_filling=True,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
                        ParameterRangeInt(name="y", type="int", size=10, ambient_index=10, ambient_size=20, start=10),
                    ],
                    check_lower_filling=True,
                ),
            ],
            1,
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(name="x", type="int", size=1, ambient_index=0, ambient_size=100, start=0),
                        ParameterRangeInt(name="y", type="int", size=20, ambient_index=0, ambient_size=20, start=0),
                    ],
                    check_lower_filling=True,
                ),
            ],
            id="true adjacency (same dimension)",
        ),
        pytest.param(
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeFloat(
                            name="x",
                            type="float",
                            size=1,
                            step=0.4,
                            start=-1.6,
                            ambient_index=1,
                            ambient_size=10,
                        ),
                        ParameterRangeFloat(
                            name="y",
                            type="float",
                            size=10,
                            step=0.4,
                            start=-2.0,
                            ambient_index=0,
                            ambient_size=10,
                        ),
                    ],
                    check_lower_filling=True,
                ),
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeFloat(
                            name="x",
                            type="float",
                            size=1,
                            step=0.4,
                            start=-2.0,
                            ambient_index=0,
                            ambient_size=10,
                        ),
                        ParameterRangeFloat(
                            name="y",
                            type="float",
                            size=10,
                            step=0.4,
                            start=-2.0,
                            ambient_index=0,
                            ambient_size=10,
                        ),
                    ],
                    check_lower_filling=True,
                ),
            ],
            0,
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeFloat(
                            name="x",
                            type="float",
                            size=2,
                            step=0.4,
                            start=-2.0,
                            ambient_index=0,
                            ambient_size=10,
                        ),
                        ParameterRangeFloat(
                            name="y",
                            type="float",
                            size=10,
                            step=0.4,
                            start=-2.0,
                            ambient_index=0,
                            ambient_size=10,
                        ),
                    ],
                    check_lower_filling=True,
                ),
            ],
            id="true adjacency (same dimension; error case found in example)",
        ),
    ],
)
def test_simplify(
    sub_spaces: list[ParameterAlignedSpace],
    target_dim: int,
    expected: list[ParameterAlignedSpace],
) -> None:
    actual = simplify(sub_spaces, target_dim)
    assert actual == expected


@pytest.mark.parametrize(
    ("segments", "expected"),
    [
        pytest.param(
            [
                FlattenSegment(0, 5),
                FlattenSegment(5, 5),
                FlattenSegment(10, 5),
            ],
            [
                FlattenSegment(0, 15),
            ],
            id="continuing 3",
        ),
        pytest.param(
            [
                FlattenSegment(0, 204),
                FlattenSegment(204, 51),
                FlattenSegment(255, 51),
                FlattenSegment(306, 51),
            ],
            [
                FlattenSegment(0, 357),
            ],
            id="continuing 4 sorted",
        ),
        pytest.param(
            [
                FlattenSegment(306, 51),
                FlattenSegment(0, 204),
                FlattenSegment(255, 51),
                FlattenSegment(204, 51),
            ],
            [
                FlattenSegment(0, 357),
            ],
            id="continuing 4 unsorted",
        ),
    ],
)
def test_simplify_simple_flatten(segments: list[FlattenSegment], expected: list[FlattenSegment]) -> None:
    actual = simplify(segments)
    assert actual == expected


@pytest.mark.parametrize(
    ("aps", "dim", "expected"),
    [
        (
            [],
            3,
            {-1: [], 0: [], 1: [], 2: []},
        ),
        (
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=1, ambient_index=0, ambient_size=10, start=0),
                        ParameterRangeInt(type="int", size=10, ambient_index=0, ambient_size=100, start=0),
                    ],
                    check_lower_filling=True,
                ),
            ],
            2,
            {
                -1: [],
                0: [],
                1: [
                    ParameterAlignedSpace(
                        axes=[
                            ParameterRangeInt(type="int", size=1, ambient_index=0, ambient_size=10, start=0),
                            ParameterRangeInt(type="int", size=10, ambient_index=0, ambient_size=100, start=0),
                        ],
                        check_lower_filling=True,
                    ),
                ],
            },
        ),
        (
            [
                ParameterAlignedSpace(
                    axes=[
                        ParameterRangeInt(type="int", size=1, ambient_index=0, ambient_size=10, start=0),
                        ParameterRangeInt(type="int", size=100, ambient_index=0, ambient_size=100, start=0),
                    ],
                    check_lower_filling=True,
                ),
            ],
            2,
            {
                -1: [],
                0: [
                    ParameterAlignedSpace(
                        axes=[
                            ParameterRangeInt(type="int", size=1, ambient_index=0, ambient_size=10, start=0),
                            ParameterRangeInt(type="int", size=100, ambient_index=0, ambient_size=100, start=0),
                        ],
                        check_lower_filling=True,
                    ),
                ],
                1: [],
            },
        ),
    ],
)
def test_remap_space(
    aps: list[ParameterAlignedSpace],
    dim: int,
    expected: dict[int, list[ParameterAlignedSpace]],
) -> None:
    actual = remap_space(aps, dim)
    assert actual == expected


@pytest.mark.parametrize(
    ("iterators", "max_num", "expected"),
    [
        (
            (range(8),),
            3,
            [(0,), (1,), (2,)],
        ),
        (
            (range(8), range(4)),
            5,
            [(0, 0), (0, 1), (0, 2), (0, 3), (1, 0)],
        ),
        (
            (itertools.count(0),),
            3,
            [(0,), (1,), (2,)],
        ),
        (
            (itertools.count(0), range(4)),
            5,
            [(0, 0), (0, 1), (0, 2), (0, 3), (1, 0)],
        ),
    ],
)
def test_infinite_product(iterators: tuple[iter, ...], max_num: int, expected: list[tuple[Any, ...]]) -> None:
    actual = []
    for item in infinite_product(*iterators):
        actual.append(item)
        if len(actual) >= max_num:
            break
    assert actual == expected


def test_infinite_product_islice() -> None:
    # noinspection PyTypeChecker
    gen = infinite_product(*(itertools.count(0), (x for x in range(8))))
    gen = itertools.islice(gen, 6, None)
    item = next(gen)
    assert item == (0, 6)
    item = next(gen)
    assert item == (0, 7)
    item = next(gen)
    assert item == (1, 0)
    item = next(gen)
    assert item == (1, 1)


def test_infinite_product_enumerate_islice() -> None:
    # noinspection PyTypeChecker
    gen = infinite_product(*(itertools.count(0), (x for x in range(8))))
    gen = enumerate(itertools.islice(gen, 6, None))
    item = next(gen)
    assert item == (0, (0, 6))
    item = next(gen)
    assert item == (1, (0, 7))
    item = next(gen)
    assert item == (2, (1, 0))
    item = next(gen)
    assert item == (3, (1, 1))



class LD2Error(Exception):
    pass


class LD2ModelTypeError(LD2Error):
    def __init__(self, type_name: str) -> None:
        super().__init__(f"Unknown type: {type_name}")


class LD2ParameterError(LD2Error):
    def __init__(self, target_param: str, error_type: str) -> None:
        super().__init__(f"Invalid parameter: {target_param}; Error type: {error_type}")


class LD2UndefinedError(LD2Error):
    def __init__(self, name: str) -> None:
        super().__init__(f"Undefined: {name}")


class LD2InvalidSpaceError(LD2Error):
    def __init__(self, msg: str) -> None:
        super().__init__(f"Invalid space: {msg}")


class LD2TypeError(LD2Error):
    def __init__(self, name: str, available_types: type | tuple[type], actual: type) -> None:
        super().__init__(f"Invalid type: {name=}, {available_types=}, {actual=}")


class LD2TableNodeServerError(LD2Error):
    pass


class LD2TableNodeClientError(LD2Error):
    pass



import pytest

from lite_dist2.curriculum_models.trial import Mapping, Trial, TrialModel, TrialStatus
from lite_dist2.value_models.aligned_space import ParameterAlignedSpaceModel
from lite_dist2.value_models.jagged_space import ParameterJaggedSpace, ParameterJaggedSpaceModel
from lite_dist2.value_models.line_segment import DummyLineSegment, LineSegmentModel
from lite_dist2.value_models.point import ResultType, ScalarValue, VectorValue
from tests.const import DT

_dummy_space = ParameterJaggedSpace(
    parameters=[(0, 1)],
    ambient_indices=[(0, 1)],
    axes_info=[DummyLineSegment(name="x", type="int", ambient_size=6, step=1)],
)


@pytest.mark.parametrize(
    ("trial", "target", "expected"),
    [
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t01",
                timestamp=DT,
                trial_status=TrialStatus.running,
                parameter_space=_dummy_space,
                result_type="scalar",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=[
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x0"),),
                        result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                    ),
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x1"),),
                        result=ScalarValue(type="scalar", value_type="int", value="0x67"),
                    ),
                ],
            ),
            ScalarValue(type="scalar", value_type="int", value="0x66"),
            None,
            id="not found: running",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t01",
                timestamp=DT,
                trial_status=TrialStatus.done,
                parameter_space=_dummy_space,
                result_type="scalar",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=None,
            ),
            ScalarValue(type="scalar", value_type="int", value="0x66"),
            None,
            id="not found: none result",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t01",
                timestamp=DT,
                trial_status=TrialStatus.done,
                parameter_space=_dummy_space,
                result_type="scalar",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=[],
            ),
            ScalarValue(type="scalar", value_type="int", value="0x66"),
            None,
            id="not found: empty result",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t01",
                timestamp=DT,
                trial_status=TrialStatus.done,
                parameter_space=_dummy_space,
                result_type="scalar",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=[
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x0"),),
                        result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                    ),
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x1"),),
                        result=ScalarValue(type="scalar", value_type="int", value="0x67"),
                    ),
                ],
            ),
            ScalarValue(type="scalar", value_type="int", value="0x66"),
            Mapping(
                params=(ScalarValue(type="scalar", value_type="int", value="0x0"),),
                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
            ),
            id="found: scalar",
        ),
        pytest.param(
            Trial(
                study_id="s01",
                trial_id="t01",
                timestamp=DT,
                trial_status=TrialStatus.done,
                parameter_space=_dummy_space,
                result_type="vector",
                result_value_type="int",
                worker_node_name="w01",
                worker_node_id="w01",
                results=[
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x0"),),
                        result=VectorValue(type="vector", value_type="int", values=["0x66", "0x2328"]),
                    ),
                    Mapping(
                        params=(ScalarValue(type="scalar", value_type="int", value="0x1"),),
                        result=VectorValue(type="vector", value_type="int", values=["0x67", "0x2329"]),
                    ),
                ],
            ),
            VectorValue(type="vector", value_type="int", values=["0x66", "0x2328"]),
            Mapping(
                params=(ScalarValue(type="scalar", value_type="int", value="0x0"),),
                result=VectorValue(type="vector", value_type="int", values=["0x66", "0x2328"]),
            ),
            id="found: vector",
        ),
    ],
)
def test_trial_find_target_value(trial: Trial, target: ResultType, expected: Mapping | None) -> None:
    actual = trial.find_target_value(target)
    assert actual == expected


@pytest.mark.parametrize(
    "model",
    [
        TrialModel(
            study_id="my_study_id0",
            trial_id="01",
            timestamp=DT,
            trial_status=TrialStatus.done,
            parameter_space=ParameterAlignedSpaceModel(
                type="aligned",
                axes=[
                    LineSegmentModel(
                        name="x",
                        type="bool",
                        size="0x2",
                        step="0x1",
                        start=False,
                        ambient_index="0x0",
                        ambient_size="0x2",
                    ),
                ],
                check_lower_filling=True,
            ),
            result_type="scalar",
            result_value_type="bool",
            worker_node_name="w01",
            worker_node_id="w01",
        ),
        TrialModel(
            study_id="my_study_id1",
            trial_id="01",
            timestamp=DT,
            trial_status=TrialStatus.done,
            parameter_space=ParameterAlignedSpaceModel(
                type="aligned",
                axes=[
                    LineSegmentModel(
                        name="x",
                        type="int",
                        size="0x2",
                        step="0x1",
                        start="0x0",
                        ambient_index="0x0",
                        ambient_size="0x2",
                    ),
                ],
                check_lower_filling=True,
            ),
            result_type="scalar",
            result_value_type="float",
            worker_node_name="w01",
            worker_node_id="w01",
            results=[
                Mapping(
                    params=(
                        ScalarValue(
                            type="scalar",
                            value_type="int",
                            value="0x1",
                            name="x",
                        ),
                    ),
                    result=ScalarValue(
                        type="scalar",
                        value_type="float",
                        value="0x1.0000000000000p+1",
                        name="r1",
                    ),
                ),
            ],
        ),
        TrialModel(
            study_id="my_study_id2",
            trial_id="02",
            timestamp=DT,
            trial_status=TrialStatus.done,
            parameter_space=ParameterJaggedSpaceModel(
                type="jagged",
                axes_info=[
                    LineSegmentModel(
                        name="x",
                        type="int",
                        size="0x1",
                        start="0x0",
                        step="0x1",
                        ambient_index="0x0",
                        ambient_size="0x1",
                        is_dummy=True,
                    ),
                ],
                parameters=[("0x1",)],
                ambient_indices=[("0x1",)],
            ),
            result_type="scalar",
            result_value_type="int",
            worker_node_name="w01",
            worker_node_id="w01",
        ),
        TrialModel(
            study_id="my_study_id3",
            trial_id="01",
            timestamp=DT,
            trial_status=TrialStatus.done,
            parameter_space=ParameterJaggedSpaceModel(
                type="jagged",
                axes_info=[
                    LineSegmentModel(
                        name="x",
                        type="int",
                        size="0x1",
                        start="0x0",
                        step="0x1",
                        ambient_index="0x0",
                        ambient_size="0x1",
                        is_dummy=True,
                    ),
                ],
                parameters=[("0x1",)],
                ambient_indices=[("0x1",)],
            ),
            result_type="scalar",
            result_value_type="int",
            worker_node_name="w01",
            worker_node_id="w01",
            results=[
                Mapping(
                    params=(
                        ScalarValue(
                            type="scalar",
                            value_type="int",
                            value="0x1",
                            name="x",
                        ),
                    ),
                    result=ScalarValue(
                        type="scalar",
                        value_type="int",
                        value="0x2",
                        name="r1",
                    ),
                ),
            ],
        ),
    ],
)
def test_trial_to_model_from_model(model: TrialModel) -> None:
    trial = Trial.from_model(model)
    reconstructed_trial = trial.to_model()
    assert model == reconstructed_trial



from __future__ import annotations

import json
import logging
import threading
import time
from typing import TYPE_CHECKING

from pydantic import BaseModel

from lite_dist2.common import publish_timestamp
from lite_dist2.config import TableConfigProvider
from lite_dist2.curriculum_models.study import Study
from lite_dist2.curriculum_models.study_portables import StudyModel, StudyStorage, StudySummary
from lite_dist2.curriculum_models.study_status import StudyStatus
from lite_dist2.expections import LD2ParameterError

if TYPE_CHECKING:
    import pathlib


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class CurriculumModel(BaseModel):
    studies: list[StudyModel]
    storages: list[StudyStorage]


class Curriculum:
    def __init__(self, studies: list[Study], storages: list[StudyStorage]) -> None:
        self.studies = studies
        self.storages = storages
        self._lock = threading.Lock()

    def get_available_study(self, retaining_capacity: set[str]) -> Study | None:
        with self._lock:
            for study in self.studies:
                if study.status == StudyStatus.running and study.required_capacity.issubset(retaining_capacity):
                    return study
            for study in self.studies:
                if study.status == StudyStatus.wait and study.required_capacity.issubset(retaining_capacity):
                    return study
        return None

    def find_study_by_id(self, study_id: str) -> Study | None:
        with self._lock:
            for study in self.studies:
                if study.study_id == study_id:
                    return study
        return None

    def try_insert_study(self, study: Study) -> bool:
        with self._lock:
            study_names = {st.name for st in self.studies if st.name is not None}
            storage_names = {st.name for st in self.storages if st.name is not None}
            if study.name is not None and (study.name in study_names or study.name in storage_names):
                return False

            self.studies.append(study)
        return True

    def to_storage_if_done(self) -> None:
        with self._lock:
            self.studies = [study for study in self.studies if not self._move_to_storage_if_done(study)]

    def _move_to_storage_if_done(self, study: Study) -> bool:
        study.update_status()
        if study.is_done():
            self.storages.append(study.to_storage())
            return True
        return False

    def pop_storage(self, study_id: str | None, name: str | None) -> StudyStorage | None:
        storages = []
        target = None
        if study_id is not None:
            for storage in self.storages:
                if storage.study_id == study_id:
                    target = storage
                    continue
                storages.append(storage)
            self.storages = storages
            return target

        if name is not None:
            for storage in self.storages:
                if storage.name == name:
                    target = storage
                    continue
                storages.append(storage)
            self.storages = storages
            return target
        p = "study_id, name"
        e = "Both are None"
        raise LD2ParameterError(p, e)

    def get_study_status(self, study_id: str | None, name: str | None) -> StudyStatus:
        if study_id is not None:
            return self._get_study_status_by_id(study_id)

        if name is not None:
            return self._get_study_status_by_name(name)

        p = "study_id, name"
        e = "Both are None"
        raise LD2ParameterError(p, e)

    def _get_study_status_by_id(self, study_id: str) -> StudyStatus:
        for study in self.studies:
            if study.study_id == study_id:
                return study.status
        for storage in self.storages:
            if storage.study_id == study_id:
                return StudyStatus.done
        return StudyStatus.not_found

    def _get_study_status_by_name(self, name: str) -> StudyStatus:
        for study in self.studies:
            if study.name == name:
                return study.status
        for storage in self.storages:
            if storage.name == name:
                return StudyStatus.done
        return StudyStatus.not_found

    def check_timeout_trial(self) -> None:
        removed_ids = []
        timeout_seconds = TableConfigProvider.get().trial_timeout_seconds
        for study in self.studies:
            removed_ids.extend(study.check_timeout_trial(publish_timestamp(), timeout_seconds))
        if len(removed_ids) > 0:
            logger.info("Outdated trials: %s", ", ".join(removed_ids))
        else:
            logger.info("No trials are outdated")

    def to_model(self) -> CurriculumModel:
        return CurriculumModel(
            studies=[study.to_model() for study in self.studies],
            storages=self.storages,
        )

    def to_summaries(self) -> list[StudySummary]:
        return [storage.to_summary() for storage in self.storages] + [study.to_summary() for study in self.studies]

    @staticmethod
    def from_model(model: CurriculumModel) -> Curriculum:
        return Curriculum(
            studies=[Study.from_model(study) for study in model.studies],
            storages=model.storages,
        )

    def save(self, curr_json_path: pathlib.Path | None = None) -> None:
        save_start_time = time.perf_counter()
        if curr_json_path is None:
            curr_json_path = TableConfigProvider.get().curriculum_path

        model = self.to_model()
        with curr_json_path.open("w", encoding="utf-8") as f:
            json.dump(model.model_dump(mode="json"), f, ensure_ascii=False)
        save_end_time = time.perf_counter()
        logger.info("Saved curriculum in %.3f msec", (save_end_time - save_start_time) / 1000)

    def cancel_study(self, study_id: str | None, name: str | None) -> bool:
        studies = []
        found = False
        if study_id is not None:
            with self._lock:
                for study in self.studies:
                    if study.study_id == study_id:
                        found = True
                        continue
                    studies.append(study)
                self.studies = studies
            return found

        if name is not None:
            with self._lock:
                for study in self.studies:
                    if study.name == name:
                        found = True
                        continue
                    studies.append(study)
                self.studies = studies
            return found
        p = "study_id, name"
        e = "Both are None"
        raise LD2ParameterError(p, e)

    @staticmethod
    def load_or_create(curr_json_path: pathlib.Path | None = None) -> Curriculum:
        load_start_time = time.perf_counter()
        if curr_json_path is None:
            curr_json_path = TableConfigProvider.get().curriculum_path

        if curr_json_path.exists():
            with curr_json_path.open("r", encoding="utf-8") as f:
                json_dict = json.load(f)
            model = CurriculumModel.model_validate(json_dict)
            return Curriculum.from_model(model)
        load_end_time = time.perf_counter()
        logger.info("Loaded curriculum in %.3f msec", (load_end_time - load_start_time) / 1000)
        return Curriculum([], [])


class CurriculumProvider:
    _CURR: Curriculum | None = None

    @classmethod
    def get(cls) -> Curriculum:
        if cls._CURR is not None:
            return cls._CURR
        cls._CURR = Curriculum.load_or_create()
        return cls._CURR

    @classmethod
    async def save_async(cls) -> None:
        if cls._CURR is None:
            return
        cls._CURR.save()

    @classmethod
    async def check_timeout(cls) -> None:
        if cls._CURR is None:
            return
        cls._CURR.check_timeout_trial()



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from lite_dist2.common import float2hex, hex2float, hex2int, int2hex
from lite_dist2.expections import LD2ParameterError, LD2UndefinedError
from lite_dist2.type_definitions import PortableValueType, PrimitiveValueType
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace
from lite_dist2.value_models.base_space import FlattenSegment, ParameterSpace
from lite_dist2.value_models.line_segment import (
    DummyLineSegment,
    LineSegment,
    LineSegmentModel,
    ParameterRangeBool,
    ParameterRangeFloat,
    ParameterRangeInt,
)
from lite_dist2.value_models.point import ParamType, ScalarValue

if TYPE_CHECKING:
    from collections.abc import Generator


class ParameterJaggedSpaceModel(BaseModel):
    type: Literal["jagged"]
    parameters: list[tuple[PortableValueType, ...]]
    ambient_indices: list[tuple[str, ...]]
    axes_info: list[LineSegmentModel]


class ParameterJaggedSpace(ParameterSpace):
    def __init__(
        self,
        parameters: list[tuple[PrimitiveValueType, ...]],
        ambient_indices: list[tuple[int, ...]],
        axes_info: list[DummyLineSegment],
    ) -> None:
        self.parameters = parameters
        self.ambient_indices = ambient_indices
        self.axes_info = axes_info

    def __eq__(self, other: object) -> bool:
        # for cache and test
        if isinstance(other, ParameterJaggedSpace):
            return (self.parameters == other.parameters) and (self.axes_info == other.axes_info)
        return False

    def __hash__(self) -> int:
        return hash((*self.parameters, *self.ambient_indices, *self.axes_info))

    def get_dim(self) -> int:
        return len(self.axes_info)

    def get_total(self) -> int:
        return len(self.parameters)

    def grid(self) -> Generator[tuple[PrimitiveValueType, ...], None, None]:
        yield from self.parameters

    def indexed_grid(self) -> Generator[tuple[tuple[int, PrimitiveValueType], ...], None, None]:
        raise NotImplementedError  # 到達しないはず??

    def value_tuple_to_param_type(self, values: tuple[PrimitiveValueType, ...]) -> ParamType:
        return tuple(
            [
                ScalarValue(type="scalar", value_type=ax.type, value=val, name=ax.name)
                for val, ax in zip(values, self.axes_info, strict=True)
            ],
        )

    def derived_by_same_ambient_space_with(self, other: ParameterSpace) -> bool:
        if isinstance(other, ParameterJaggedSpace):
            return self.axes_info == other.axes_info
        return False

    def to_aligned_list(self) -> list[ParameterAlignedSpace]:
        segment_types: list[type[LineSegment]] = []
        for dummy in self.axes_info:
            match dummy.type:
                case "bool":
                    segment_types.append(ParameterRangeBool)
                case "int":
                    segment_types.append(ParameterRangeInt)
                case "float":
                    segment_types.append(ParameterRangeFloat)
                case _:
                    raise LD2UndefinedError(dummy.type)

        space_by_line = defaultdict(list)
        for ambient_index, param in zip(self.ambient_indices, self.parameters, strict=True):
            space_by_line[ambient_index[1:]].append(
                ParameterAlignedSpace(
                    axes=[
                        lst(
                            type=axis_info.type,
                            name=axis_info.name,
                            start=p,
                            size=1,
                            step=axis_info.step,
                            ambient_index=amb_idx,
                            ambient_size=axis_info.ambient_size,
                        )
                        for lst, axis_info, p, amb_idx in zip(
                            segment_types,
                            self.axes_info,
                            param,
                            ambient_index,
                            strict=True,
                        )
                    ],
                    check_lower_filling=True,
                ),
            )
        spaces = []
        for v in space_by_line.values():
            spaces.extend(v)
        return spaces

    def lower_element_num_by_dim(self) -> tuple[int, ...]:
        return self.get_lower_element_num_by_dim([axis.ambient_size for axis in self.axes_info])

    def get_flatten_ambient_start_and_size_list(self) -> list[FlattenSegment]:
        lower_element_num_by_dim = self.lower_element_num_by_dim()
        flatten_segments = []
        for amb_idx in self.ambient_indices:
            flatten_index = sum(d * lower for d, lower in zip(amb_idx, lower_element_num_by_dim, strict=True))
            flatten_segments.append(FlattenSegment(flatten_index, 1))
        return flatten_segments

    def to_model(self) -> ParameterJaggedSpaceModel:
        return ParameterJaggedSpaceModel(
            type="jagged",
            parameters=[tuple(self._primitive_to_portable(p) for p in primitive) for primitive in self.parameters],
            ambient_indices=[tuple(int2hex(idx) for idx in amb_idx) for amb_idx in self.ambient_indices],
            axes_info=[axis.to_model() for axis in self.axes_info],
        )

    @staticmethod
    def from_model(model: ParameterJaggedSpaceModel) -> ParameterJaggedSpace:
        axes_info = []
        for axis in model.axes_info:
            if not axis.is_dummy:
                param = f"{LineSegmentModel.__name__}.type"
                msg = f"An axis of {ParameterJaggedSpace.__name__} is only allowed dummy axis."
                raise LD2ParameterError(param, msg)
            axes_info.append(DummyLineSegment.from_model(axis))

        return ParameterJaggedSpace(
            parameters=[
                tuple(ParameterJaggedSpace._portable_to_primitive(p) for p in port) for port in model.parameters
            ],
            ambient_indices=[tuple(hex2int(idx) for idx in amb_idx) for amb_idx in model.ambient_indices],
            axes_info=axes_info,
        )

    @staticmethod
    def _primitive_to_portable(primitive: PrimitiveValueType) -> PortableValueType:
        match primitive:
            case bool():
                return primitive
            case int():
                return int2hex(primitive)
            case float():
                return float2hex(primitive)
            case _:
                raise LD2UndefinedError(str(type(primitive)))

    @staticmethod
    def _portable_to_primitive(portable: PortableValueType) -> PrimitiveValueType:
        match portable:
            case bool():
                return portable
            case str() if "p" in portable:
                return hex2float(portable)
            case str() if portable.startswith("0x"):
                return hex2int(portable)
            case _:
                raise LD2UndefinedError(str(type(portable)))



from __future__ import annotations

import logging
import time
import uuid
from typing import TYPE_CHECKING, Annotated

from lite_dist2.expections import LD2TableNodeServerError
from lite_dist2.worker_node.table_node_client import TableNodeClient

if TYPE_CHECKING:
    from multiprocessing.pool import Pool

    from lite_dist2.config import WorkerConfig
    from lite_dist2.worker_node.trial_runner import BaseTrialRunner


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Worker:
    def __init__(
        self,
        trial_runner: Annotated[BaseTrialRunner, "Runner for executing trials"],
        ip: Annotated[str, "IP address of the table node server"],
        port: Annotated[int | str, "Port number  of the table node server"],
        config: Annotated[WorkerConfig, "Configuration of  the worker node"],
        pool: Annotated[
            Pool | None,
            "Process pool for parallel execution. Ignored except `SemiAutoMPTrialRunner`.",
        ] = None,
    ) -> None:
        self.trial_runner = trial_runner
        self.client = TableNodeClient(ip, port)
        self.pool = pool
        self.config = config
        self.id = str(uuid.uuid1())

    def start(self, stop_at_no_trial: bool = False, *args: object, **kwargs: object) -> None:
        if not self.client.ping():
            msg = "Table node server not responding"
            raise LD2TableNodeServerError(msg)

        while True:
            has_next = self._step(*args, **kwargs)
            if (not has_next) and stop_at_no_trial:
                logger.info("No trial. Stop worker after saving.")
                self.client.save()
                return
            if not has_next:
                logger.info("No trial. Waiting %d seconds...", self.config.wait_seconds_on_no_trial)
                time.sleep(self.config.wait_seconds_on_no_trial)

    def _step(self, *args: object, **kwargs: object) -> bool:
        trial = self.client.reserve_trial(
            self.id,
            self.config.name,
            self.config.max_size,
            self.config.retaining_capacity,
            self.config.table_node_request_timeout_seconds,
        )
        if trial is None:
            return False

        done_trial = self.trial_runner.run(trial, self.config, self.pool, *args, **kwargs)
        self.client.register_trial(done_trial, self.config.table_node_request_timeout_seconds)
        return True



import pytest

from lite_dist2.curriculum_models.trial import Mapping, Trial, TrialStatus
from lite_dist2.curriculum_models.trial_table import TrialTable
from lite_dist2.study_strategies.base_study_strategy import StudyStrategyParam
from lite_dist2.study_strategies.find_exact_study_strategy import FindExactStudyStrategy
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace
from lite_dist2.value_models.line_segment import ParameterRangeInt
from lite_dist2.value_models.point import ResultType, ScalarValue
from tests.const import DT

_DUMMY_PARAMETER_SPACE = ParameterAlignedSpace(
    axes=[
        ParameterRangeInt(name="x", type="int", size=2, start=0, ambient_size=2, ambient_index=0),
        ParameterRangeInt(name="y", type="int", size=2, start=0, ambient_size=2, ambient_index=0),
    ],
    check_lower_filling=True,
)
_DUMMY_APS = {-1: [], 0: [], 1: []}
_TRIAL_ARGS = {
    "study_id": "s01",
    "timestamp": DT,
    "parameter_space": _DUMMY_PARAMETER_SPACE,
    "result_type": "scalar",
    "result_value_type": "int",
    "worker_node_name": "w01",
    "worker_node_id": "w01",
}


@pytest.mark.parametrize(
    ("trial_table", "target_value", "expected"),
    [
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space=_DUMMY_APS,
            ),
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            False,
            id="Empty: False",
        ),
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        trial_id="t01",
                        trial_status=TrialStatus.done,
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                            ),
                        ],
                        **_TRIAL_ARGS,
                    ),
                ],
                aggregated_parameter_space=_DUMMY_APS,
            ),
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            False,
            id="Not found: False",
        ),
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        trial_id="t01",
                        trial_status=TrialStatus.done,
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                        ],
                        **_TRIAL_ARGS,
                    ),
                ],
                aggregated_parameter_space=_DUMMY_APS,
            ),
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            True,
            id="Found: True",
        ),
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        trial_id="t01",
                        trial_status=TrialStatus.running,
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                        ],
                        **_TRIAL_ARGS,
                    ),
                ],
                aggregated_parameter_space=_DUMMY_APS,
            ),
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            False,
            id="Found but running: False",
        ),
    ],
)
def test_find_exact_study_strategy_is_done(
    trial_table: TrialTable,
    target_value: ResultType,
    expected: bool,
) -> None:
    strategy = FindExactStudyStrategy(StudyStrategyParam(target_value=target_value))
    actual = strategy.is_done(trial_table, _DUMMY_PARAMETER_SPACE)
    assert actual == expected


@pytest.mark.parametrize(
    ("trial_table", "target_value", "expected"),
    [
        pytest.param(
            TrialTable(
                trials=[],
                aggregated_parameter_space=_DUMMY_APS,
            ),
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            [],
            id="Empty",
        ),
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        trial_id="t01",
                        trial_status=TrialStatus.done,
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x66"),
                            ),
                        ],
                        **_TRIAL_ARGS,
                    ),
                ],
                aggregated_parameter_space=_DUMMY_APS,
            ),
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            [],
            id="Not found",
        ),
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        trial_id="t01",
                        trial_status=TrialStatus.done,
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                        ],
                        **_TRIAL_ARGS,
                    ),
                ],
                aggregated_parameter_space=_DUMMY_APS,
            ),
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            [
                Mapping(
                    params=(
                        ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                        ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                    ),
                    result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                ),
            ],
            id="Found",
        ),
        pytest.param(
            TrialTable(
                trials=[
                    Trial(
                        trial_id="t01",
                        trial_status=TrialStatus.running,
                        results=[
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x64"),
                            ),
                            Mapping(
                                params=(
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                    ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                ),
                                result=ScalarValue(type="scalar", value_type="int", value="0x65"),
                            ),
                        ],
                        **_TRIAL_ARGS,
                    ),
                ],
                aggregated_parameter_space=_DUMMY_APS,
            ),
            ScalarValue(type="scalar", value_type="int", value="0x64"),
            [],
            id="Found but running",
        ),
    ],
)
def test_find_exact_study_strategy_extract_mapping(
    trial_table: TrialTable,
    target_value: ResultType,
    expected: list[Mapping],
) -> None:
    strategy = FindExactStudyStrategy(StudyStrategyParam(target_value=target_value))
    actual = strategy.extract_mappings(trial_table)
    assert actual == expected



from __future__ import annotations

import threading
from typing import TYPE_CHECKING, Literal

from lite_dist2.common import int2hex, publish_timestamp
from lite_dist2.curriculum_models.study_portables import StudyModel, StudyStorage, StudySummary
from lite_dist2.curriculum_models.study_status import StudyStatus
from lite_dist2.curriculum_models.trial import Trial, TrialStatus
from lite_dist2.curriculum_models.trial_table import TrialTable
from lite_dist2.expections import LD2ModelTypeError
from lite_dist2.study_strategies.all_calculation_study_strategy import AllCalculationStudyStrategy
from lite_dist2.study_strategies.find_exact_study_strategy import FindExactStudyStrategy
from lite_dist2.suggest_strategies import SequentialSuggestStrategy, SuggestStrategyModel
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace

if TYPE_CHECKING:
    from datetime import datetime

    from lite_dist2.study_strategies import BaseStudyStrategy, StudyStrategyModel
    from lite_dist2.suggest_strategies import BaseSuggestStrategy


class Study:
    def __init__(
        self,
        study_id: str,
        name: str | None,
        required_capacity: set[str],
        status: StudyStatus,
        registered_timestamp: datetime,
        study_strategy: BaseStudyStrategy,
        suggest_strategy: BaseSuggestStrategy,
        parameter_space: ParameterAlignedSpace,
        result_type: Literal["scalar", "vector"],
        result_value_type: Literal["bool", "int", "float"],
        trial_table: TrialTable,
    ) -> None:
        self.study_id = study_id
        self.name = name or self.study_id
        self.required_capacity = required_capacity
        self.status = status
        self.registered_timestamp = registered_timestamp
        self.study_strategy = study_strategy
        self.suggest_strategy = suggest_strategy
        self.parameter_space = parameter_space
        self.result_type = result_type
        self.result_value_type = result_value_type
        self.trial_table = trial_table

        self._table_lock = threading.Lock()

    def update_status(self) -> None:
        if self.is_done():
            self.status = StudyStatus.done
            return

    def is_done(self) -> bool:
        return self.study_strategy.is_done(self.trial_table, self.parameter_space)

    def suggest_next_trial(
        self,
        num: int | None,
        worker_node_name: str | None,
        worker_node_id: str,
    ) -> Trial | None:
        with self._table_lock:
            self.status = StudyStatus.running

            parameter_sub_space = self.suggest_strategy.suggest(self.trial_table, num)
            if parameter_sub_space is None:
                return None

            trial = Trial(
                study_id=self.study_id,
                trial_id=self._publish_trial_id(),
                timestamp=publish_timestamp(),
                trial_status=TrialStatus.running,
                parameter_space=parameter_sub_space,
                result_type=self.result_type,
                result_value_type=self.result_value_type,
                worker_node_name=worker_node_name,
                worker_node_id=worker_node_id,
            )
            self.trial_table.register(trial)
        if self.trial_table.is_not_defined_aps():
            self.trial_table.init_aps(trial)
        return trial

    def receipt_trial(self, trial: Trial) -> None:
        with self._table_lock:
            self.trial_table.receipt_trial_result(trial.trial_id, trial.result, trial.worker_node_id)
            self.trial_table.simplify_aps()

    def check_timeout_trial(self, now: datetime, timeout_seconds: int) -> list[str]:
        return self.trial_table.check_timeout_trial(now, timeout_seconds)

    def to_storage(self) -> StudyStorage:
        return StudyStorage(
            study_id=self.study_id,
            name=self.name,
            required_capacity=self.required_capacity,
            registered_timestamp=self.registered_timestamp,
            study_strategy=self.study_strategy.to_model(),
            suggest_strategy=self.suggest_strategy.to_model(),
            parameter_space=self.parameter_space.to_model(),
            done_timestamp=publish_timestamp(),
            result_type=self.result_type,
            result_value_type=self.result_value_type,
            results=self.study_strategy.extract_mappings(self.trial_table),
            done_grids=self.trial_table.count_grid(),
        )

    def to_summary(self) -> StudySummary:
        done_grids = sum(trial.parameter_space.get_total() for trial in self.trial_table.trials)
        return StudySummary(
            name=self.name,
            study_id=self.study_id,
            required_capacity=self.required_capacity,
            status=self.status,
            registered_timestamp=self.registered_timestamp,
            study_strategy=self.study_strategy.to_model(),
            suggest_strategy=self.suggest_strategy.to_model(),
            parameter_space=self.parameter_space.to_model(),
            result_type=self.result_type,
            result_value_type=self.result_value_type,
            total_grids=self.parameter_space.total,
            done_grids=done_grids,
        )

    def to_model(self) -> StudyModel:
        return StudyModel(
            study_id=self.study_id,
            name=self.name,
            required_capacity=self.required_capacity,
            status=self.status,
            registered_timestamp=self.registered_timestamp,
            study_strategy=self.study_strategy.to_model(),
            suggest_strategy=self.suggest_strategy.to_model(),
            parameter_space=self.parameter_space.to_model(),
            result_type=self.result_type,
            result_value_type=self.result_value_type,
            trial_table=self.trial_table.to_model(),
        )

    def _publish_trial_id(self) -> str:
        return f"{self.study_id}-{int2hex(self.trial_table.count_trial())}"

    @staticmethod
    def _create_study_strategy(model: StudyStrategyModel) -> BaseStudyStrategy:
        match model.type:
            case "all_calculation":
                return AllCalculationStudyStrategy(model.study_strategy_param)
            case "find_exact":
                return FindExactStudyStrategy(model.study_strategy_param)
            case "minimize":
                raise NotImplementedError
            case _:
                raise LD2ModelTypeError(model.type)

    @staticmethod
    def _create_suggest_strategy(model: SuggestStrategyModel, space: ParameterAlignedSpace) -> BaseSuggestStrategy:
        match model.type:
            case "sequential":
                return SequentialSuggestStrategy(model.suggest_strategy_param, space)
            case "random":
                raise NotImplementedError
            case "designated":
                raise NotImplementedError
            case _:
                raise LD2ModelTypeError(model.type)

    @staticmethod
    def from_model(study_model: StudyModel) -> Study:
        parameter_space = ParameterAlignedSpace.from_model(study_model.parameter_space)
        return Study(
            study_id=study_model.study_id,
            name=study_model.name,
            required_capacity=study_model.required_capacity,
            status=study_model.status,
            registered_timestamp=study_model.registered_timestamp,
            study_strategy=Study._create_study_strategy(study_model.study_strategy),
            suggest_strategy=Study._create_suggest_strategy(study_model.suggest_strategy, parameter_space),
            parameter_space=parameter_space,
            result_type=study_model.result_type,
            result_value_type=study_model.result_value_type,
            trial_table=TrialTable.from_model(study_model.trial_table),
        )



from typing import Literal

import pytest

from lite_dist2 import common
from lite_dist2.type_definitions import PortableValueType, PrimitiveValueType


@pytest.mark.parametrize(
    ("hex_str", "expected"),
    [
        ("0x0", 0),
        ("0x01", 1),
        ("0xa", 10),
        ("0xa54f", 42319),
        ("-0x1", -1),
    ],
)
def test_hex2int(hex_str: str, expected: int) -> None:
    assert common.hex2int(hex_str) == expected


@pytest.mark.parametrize(
    ("int_val", "expected"),
    [
        (0, "0x0"),
        (10, "0xa"),
        (42319, "0xa54f"),
        (-1, "-0x1"),
    ],
)
def test_int2hex(int_val: int, expected: str) -> None:
    assert common.int2hex(int_val) == expected


@pytest.mark.parametrize(
    ("hex_str", "expected"),
    [
        ("0x0.0p+0", 0.0),
        ("-0x0.0p+0", -0.0),
        ("0x1.999999999999ap-4", 0.1),
    ],
)
def test_hex2float(hex_str: str, expected: float) -> None:
    assert common.hex2float(hex_str) == expected


@pytest.mark.parametrize(
    ("float_val", "expected"),
    [
        (0.0, "0x0.0p+0"),
        (-0.0, "-0x0.0p+0"),
        (0.1, "0x1.999999999999ap-4"),
    ],
)
def test_float2hex(float_val: float, expected: str) -> None:
    assert common.float2hex(float_val) == expected


@pytest.mark.parametrize(
    ("type_name", "value", "expected"),
    [
        ("bool", False, False),
        ("int", "0x2", 2),
        ("float", "0x1.999999999999ap-4", 0.1),
    ],
)
def test_numerize(
    type_name: Literal["bool", "int", "float"],
    value: PortableValueType,
    expected: PrimitiveValueType,
) -> None:
    actual = common.numerize(type_name, value)
    assert actual == expected


@pytest.mark.parametrize(
    ("type_name", "value", "expected"),
    [
        ("bool", False, False),
        ("int", 2, "0x2"),
        ("float", 0.1, "0x1.999999999999ap-4"),
    ],
)
def test_portablize(
    type_name: Literal["bool", "int", "float"],
    value: PrimitiveValueType,
    expected: PortableValueType,
) -> None:
    actual = common.portablize(type_name, value)
    assert actual == expected



from __future__ import annotations

import itertools
from functools import reduce
from typing import TYPE_CHECKING

from lite_dist2.interfaces import Mergeable

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    from lite_dist2.value_models.aligned_space import ParameterAlignedSpace


def simplify[T: Mergeable](mergeables: list[T], *args: object) -> list[T]:
    new_aps: list[Mergeable] = []
    mergeables_duplicated_group: dict[int, set[int]] = {}
    sub_space_num = len(mergeables)
    mergeables = sorted(mergeables, key=lambda spc: spc.get_start_index(*args))

    for i in range(sub_space_num):
        for j in range(i + 1, sub_space_num):
            if mergeables[i].can_merge(mergeables[j], *args):
                if i in mergeables_duplicated_group:
                    mergeables_duplicated_group[i].add(j)
                else:
                    mergeables_duplicated_group[i] = {j}

    mergeable_group: list[set[int]] = []
    for i, mergeable_to_i in mergeables_duplicated_group.items():
        grouplet = mergeable_to_i.union({i})
        for g in range(len(mergeable_group)):
            intersection = mergeable_group[g].intersection(grouplet)
            if len(intersection) > 0:
                mergeable_group[g] = mergeable_group[g].union(grouplet)
                break
        else:
            mergeable_group.append(grouplet)

    not_mergeables = set(range(sub_space_num)) - reduce(lambda x, y: x.union(y), mergeable_group, set())
    new_aps.extend([mergeables[i] for i in not_mergeables])

    for group_index_set in mergeable_group:
        group_space_list = sorted([mergeables[i] for i in group_index_set], key=lambda spc: spc.get_start_index(*args))
        merged = group_space_list[0]
        for space in group_space_list[1:]:
            merged = merged.merge(space, *args)
        new_aps.append(merged)

    return sorted(new_aps, key=lambda spc: spc.get_start_index(*args))


def remap_space(aps: list[ParameterAlignedSpace], dim: int) -> dict[int, list[ParameterAlignedSpace]]:
    remapped = {i: [] for i in range(-1, dim)}

    for space in aps:
        universal_dim = space.get_lower_not_universal_dim()
        remapped[universal_dim].append(space)
    return remapped


def infinite_product[T](*iterators: tuple[Iterable[T], ...]) -> Generator[tuple[T, ...], None, None]:
    if len(iterators) == 1:
        for infinite_element in iterators[0]:
            yield (infinite_element,)
    else:
        lowers_original = list(iterators[1:])
        for infinite_element in iterators[0]:
            lowers = []
            _lowers_original = []
            for lower_original in lowers_original:
                _lower_original, lower = itertools.tee(lower_original)
                _lowers_original.append(_lower_original)
                lowers.append(lower)

            for lower in itertools.product(*lowers):
                yield infinite_element, *lower
            lowers_original = _lowers_original



from __future__ import annotations

import abc
import functools
from multiprocessing.pool import Pool
from typing import TYPE_CHECKING, Any

import tqdm

if TYPE_CHECKING:
    from collections.abc import Iterator

    from lite_dist2.config import WorkerConfig
    from lite_dist2.curriculum_models.trial import Trial
    from lite_dist2.type_definitions import RawParamType, RawResultType
    from lite_dist2.value_models.base_space import ParameterSpace


class BaseTrialRunner(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def func(self, parameters: RawParamType, *args: object, **kwargs: object) -> RawResultType:
        pass

    @abc.abstractmethod
    def wrap_func(
        self,
        parameter_space: ParameterSpace,
        config: WorkerConfig,
        pool: Pool | None = None,
        *args: object,
        **kwargs: object,
    ) -> list[tuple[RawParamType, RawResultType]]:
        pass

    def parameter_pass_func(
        self,
        parameters: RawParamType,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> tuple[RawParamType, RawResultType]:
        return parameters, self.func(parameters, *args, **kwargs)

    def run(
        self,
        trial: Trial,
        config: WorkerConfig,
        pool: Pool | None = None,
        *args: object,
        **kwargs: object,
    ) -> Trial:
        raw_mappings = self.wrap_func(trial.parameter_space, config, pool, *args, **kwargs)
        mappings = trial.convert_mappings_from(raw_mappings)
        trial.set_result(mappings)
        return trial


class AutoMPTrialRunner(BaseTrialRunner, metaclass=abc.ABCMeta):
    def wrap_func(
        self,
        parameter_space: ParameterSpace,
        config: WorkerConfig,
        _: Pool | None = None,
        *args: object,
        **kwargs: object,
    ) -> list[tuple[RawParamType, RawResultType]]:
        raw_mappings: list[tuple[RawParamType, RawResultType]] = []
        total = parameter_space.get_total()
        tqdm_kwargs = {"total": total, "disable": config.disable_function_progress_bar}
        if config.process_num is None or config.process_num > 1:
            parameter_pass_func = functools.partial(self.parameter_pass_func, args=args, kwargs=kwargs)
            with Pool(processes=config.process_num) as pool, tqdm.tqdm(**tqdm_kwargs) as p_bar:
                for arg_tuple, result_iter in pool.imap_unordered(
                    func=parameter_pass_func,
                    iterable=parameter_space.grid(),
                    chunksize=config.chunk_size,
                ):
                    raw_mappings.append((arg_tuple, result_iter))
                    p_bar.update(1)
            return raw_mappings
        return [
            self.parameter_pass_func(arg_tuple, args, kwargs)
            for arg_tuple in tqdm.tqdm(parameter_space.grid(), **tqdm_kwargs)
        ]


class SemiAutoMPTrialRunner(BaseTrialRunner, metaclass=abc.ABCMeta):
    def wrap_func(
        self,
        parameter_space: ParameterSpace,
        config: WorkerConfig,
        pool: Pool | None = None,
        *args: object,
        **kwargs: object,
    ) -> list[tuple[RawParamType, RawResultType]]:
        raw_mappings: list[tuple[RawParamType, RawResultType]] = []
        total = parameter_space.get_total()
        tqdm_kwargs = {"total": total, "disable": config.disable_function_progress_bar}
        if pool is not None:
            parameter_pass_func = functools.partial(self.parameter_pass_func, args=args, kwargs=kwargs)
            with tqdm.tqdm(**tqdm_kwargs) as p_bar:
                for arg_tuple, result_iter in pool.imap_unordered(
                    func=parameter_pass_func,
                    iterable=parameter_space.grid(),
                    chunksize=config.chunk_size,
                ):
                    raw_mappings.append((arg_tuple, result_iter))
                    p_bar.update(1)
            return raw_mappings
        return [
            self.parameter_pass_func(arg_tuple, args, kwargs)
            for arg_tuple in tqdm.tqdm(parameter_space.grid(), **tqdm_kwargs)
        ]


class ManualMPTrialRunner(BaseTrialRunner, metaclass=abc.ABCMeta):
    def func(self, *parameters: RawParamType) -> tuple[RawParamType, RawResultType]:
        pass

    @abc.abstractmethod
    def batch_func(
        self,
        raw_params: Iterator[RawParamType],
        config: WorkerConfig,
        *args: object,
        **kwargs: object,
    ) -> list[tuple[RawParamType, RawResultType]]:
        pass

    def wrap_func(
        self,
        parameter_space: ParameterSpace,
        config: WorkerConfig,
        _: Pool | None = None,
        *args: object,
        **kwargs: object,
    ) -> list[tuple[RawParamType, RawResultType]]:
        return self.batch_func(parameter_space.grid(), config, *args, **kwargs)



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from lite_dist2.value_models.point import ResultType

if TYPE_CHECKING:
    from lite_dist2.curriculum_models.trial import Mapping
    from lite_dist2.curriculum_models.trial_table import TrialTable
    from lite_dist2.value_models.aligned_space import ParameterAlignedSpace


class StudyStrategyParam(BaseModel):
    target_value: ResultType


class StudyStrategyModel(BaseModel):
    type: Literal["all_calculation", "find_exact", "minimize"]
    study_strategy_param: StudyStrategyParam | None


class BaseStudyStrategy(metaclass=abc.ABCMeta):
    def __init__(self, study_strategy_param: StudyStrategyParam | None) -> None:
        self.study_strategy_param = study_strategy_param

    @abc.abstractmethod
    def is_done(self, trial_table: TrialTable, parameter_space: ParameterAlignedSpace) -> bool:
        pass

    @abc.abstractmethod
    def extract_mappings(self, trial_table: TrialTable) -> list[Mapping]:
        pass

    @abc.abstractmethod
    def to_model(self) -> StudyStrategyModel:
        pass



from typing import Any, Literal

import pytest

from lite_dist2.expections import LD2ModelTypeError
from lite_dist2.value_models.point import ScalarValue, VectorValue


@pytest.mark.parametrize(
    ("json_dict", "expected"),
    [
        (
            {"type": "scalar", "value_type": "bool", "value": True, "name": "some flag"},
            ScalarValue(type="scalar", value_type="bool", value=True, name="some flag"),
        ),
        (
            {"type": "scalar", "value_type": "int", "value": "f23", "name": "some int"},
            ScalarValue(type="scalar", value_type="int", value="f23", name="some int"),
        ),
        (
            {"type": "scalar", "value_type": "float", "value": "0x1.999999999999ap-4", "name": "some float"},
            ScalarValue(type="scalar", value_type="float", value="0x1.999999999999ap-4", name="some float"),
        ),
    ],
)
def test_scalar_value_deserialize(json_dict: dict[str, Any], expected: ScalarValue) -> None:
    actual = ScalarValue.model_validate(json_dict)
    assert actual == expected


@pytest.mark.parametrize(
    ("scalar_value", "expected"),
    [
        (ScalarValue(type="scalar", value_type="bool", value=True), True),
        (ScalarValue(type="scalar", value_type="int", value="f23"), 3875),
        (ScalarValue(type="scalar", value_type="float", value="0x1.999999999999ap-4"), 0.1),
    ],
)
def test_scalar_value_numerize(scalar_value: ScalarValue, expected: bool | float) -> None:
    actual = scalar_value.numerize()
    assert actual == expected


def test_scalar_value_numerize_value_type_error() -> None:
    scalar = ScalarValue(type="scalar", value_type="bool", value=True)
    # noinspection PyTypeChecker
    scalar.value_type = "invalid"
    with pytest.raises(LD2ModelTypeError, match=r"Unknown\stype:\s"):
        _ = scalar.numerize()


@pytest.mark.parametrize(
    ("raw_result_value", "value_type", "name", "expected"),
    [
        (True, "bool", "flag", ScalarValue(type="scalar", value_type="bool", value=True, name="flag")),
        (3875, "int", "num", ScalarValue(type="scalar", value_type="int", value="0xf23", name="num")),
        (0.1, "float", None, ScalarValue(type="scalar", value_type="float", value="0x1.999999999999ap-4")),
    ],
)
def test_scalar_value_create_from_numeric(
    raw_result_value: bool | float,
    value_type: Literal["bool", "int", "float"],
    name: str | None,
    expected: ScalarValue,
) -> None:
    actual = ScalarValue.create_from_numeric(raw_result_value, value_type, name)
    assert actual == expected


def test_scalar_value_create_from_numeric_value_type_error() -> None:
    with pytest.raises(LD2ModelTypeError, match=r"Unknown\stype:\s"):
        # noinspection PyTypeChecker
        _ = ScalarValue.create_from_numeric(raw_result_value=True, value_type="invalid")


@pytest.mark.parametrize(
    ("json_dict", "expected"),
    [
        (
            {"type": "vector", "value_type": "bool", "values": [True], "name": "some flag"},
            VectorValue(type="vector", value_type="bool", values=[True], name="some flag"),
        ),
        (
            {"type": "vector", "value_type": "int", "values": ["0xf23"], "name": "some int"},
            VectorValue(type="vector", value_type="int", values=["0xf23"], name="some int"),
        ),
        (
            {"type": "vector", "value_type": "float", "values": ["0x1.999999999999ap-4"], "name": "some float"},
            VectorValue(type="vector", value_type="float", values=["0x1.999999999999ap-4"], name="some float"),
        ),
    ],
)
def test_vector_value_deserialize(json_dict: dict[str, Any], expected: VectorValue) -> None:
    actual = VectorValue.model_validate(json_dict)
    assert actual == expected


@pytest.mark.parametrize(
    ("vector_value", "expected"),
    [
        (VectorValue(type="vector", value_type="bool", values=[True]), [True]),
        (VectorValue(type="vector", value_type="int", values=["f23"]), [3875]),
        (VectorValue(type="vector", value_type="float", values=["0x1.999999999999ap-4"]), [0.1]),
    ],
)
def test_vector_value_numerize(vector_value: VectorValue, expected: list[bool | int | float]) -> None:
    actual = vector_value.numerize()
    assert actual == expected


def test_vector_value_numerize_value_type_error() -> None:
    vector = VectorValue(type="vector", value_type="bool", values=[True])
    # noinspection PyTypeChecker
    vector.value_type = "invalid"
    with pytest.raises(LD2ModelTypeError, match=r"Unknown\stype:\s"):
        _ = vector.numerize()


@pytest.mark.parametrize(
    ("raw_result_value", "value_type", "name", "expected"),
    [
        ([True], "bool", "flag", VectorValue(type="vector", value_type="bool", values=[True], name="flag")),
        ([3875], "int", "num", VectorValue(type="vector", value_type="int", values=["0xf23"], name="num")),
        ([0.1], "float", None, VectorValue(type="vector", value_type="float", values=["0x1.999999999999ap-4"])),
    ],
)
def test_vector_value_create_from_numeric(
    raw_result_value: list[bool | int | float],
    value_type: Literal["bool", "int", "float"],
    name: str | None,
    expected: VectorValue,
) -> None:
    actual = VectorValue.create_from_numeric(raw_result_value, value_type, name)
    assert actual == expected


def test_vector_value_create_from_numeric_value_type_error() -> None:
    with pytest.raises(LD2ModelTypeError, match=r"Unknown\stype:\s"):
        # noinspection PyTypeChecker
        _ = VectorValue.create_from_numeric(raw_result_value=[True], value_type="invalid")



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, Any

from lite_dist2.expections import LD2InvalidSpaceError, LD2ParameterError
from lite_dist2.suggest_strategies import BaseSuggestStrategy
from lite_dist2.suggest_strategies.base_suggest_strategy import SuggestStrategyModel
from lite_dist2.value_models.jagged_space import ParameterJaggedSpace

if TYPE_CHECKING:
    from collections.abc import Generator

    from lite_dist2.curriculum_models.trial_table import TrialTable
    from lite_dist2.suggest_strategies.base_suggest_strategy import SuggestStrategyParam
    from lite_dist2.value_models.aligned_space import ParameterAlignedSpace
    from lite_dist2.value_models.base_space import ParameterSpace


class SequentialSuggestStrategy(BaseSuggestStrategy):
    def __init__(
        self,
        suggest_parameter: SuggestStrategyParam,
        parameter_space: ParameterAlignedSpace,
    ) -> None:
        super().__init__(suggest_parameter, parameter_space)
        self.strict_aligned = self.suggest_parameter.strict_aligned

    def suggest(self, trial_table: TrialTable, max_num: int) -> ParameterSpace | None:
        least_seg = trial_table.find_least_division(self.parameter_space.total)
        if least_seg.size == 0:
            return None
        capped_max_num = self._nullable_min(least_seg.size, max_num)
        start = least_seg.start

        if self.strict_aligned or self.parameter_space.dim == 1:
            return self._aligned_suggest(start, capped_max_num)
        return self._jagged_suggest(start, capped_max_num)

    def _aligned_suggest(self, start: int, max_num: int) -> ParameterAlignedSpace:
        if self.parameter_space.is_infinite():
            available_next, infinite_flag = self._generate_available_next_infinite(start)
            if infinite_flag:
                max_available_gen = self._infinite_available_generator(
                    available_next,
                    self.parameter_space.lower_element_num_by_dim()[0],
                )
                infinite_available_next = set()
                for _next_index in max_available_gen:
                    infinite_available_next.add(_next_index)
                    if _next_index - start >= max_num:
                        break
                max_available_next = max(infinite_available_next)
            else:
                max_available_next = self._calc_max_available_next(available_next, start, max_num)

        else:
            available_next = self._generate_available_next_finite(start)
            max_available_next = self._calc_max_available_next(available_next, start, max_num)

        start_loom = self.parameter_space.loom_by_flatten_index(
            start,
            self.parameter_space.lower_element_num_by_dim(),
        )
        end_loom = self.parameter_space.loom_by_flatten_index(
            max_available_next - 1,
            self.parameter_space.lower_element_num_by_dim(),
        )
        start_and_sizes = [(s, e - s + 1) for s, e in zip(start_loom, end_loom, strict=True)]
        return self.parameter_space.slice(start_and_sizes)

    def _jagged_suggest(self, start: int, max_num: int) -> ParameterJaggedSpace | None:
        gen = itertools.islice(self.parameter_space.indexed_grid(), start, None)
        parameters = []
        ambient_indices = []
        for i, ai_param in enumerate(gen):
            ambient_index = tuple(aip[0] for aip in ai_param)
            param = tuple(aip[1] for aip in ai_param)

            parameters.append(param)
            ambient_indices.append(ambient_index)
            if i + 1 >= max_num:
                break

        if len(parameters) == 0:
            return None
        return ParameterJaggedSpace(parameters, ambient_indices, self.parameter_space.dummy_info)

    @staticmethod
    def _calc_max_available_next(available_next: tuple[int, ...], start: int, max_num: int) -> int:
        available = list(filter(lambda next_index: next_index - start <= max_num, available_next))
        if len(available) == 0:
            msg = "No available"
            raise LD2InvalidSpaceError(msg)
        return max(available)

    @staticmethod
    def _nullable_min(a: int | None, b: int | None) -> int:
        if a is not None and b is not None:
            return min(a, b)
        if a is not None:
            return a
        if b is not None:
            return b
        target_param = "a, b"
        error_type = "both is None"
        raise LD2ParameterError(target_param, error_type)

    def _generate_available_next_finite(self, flatten_index: int) -> tuple[int, ...]:
        dims = self.parameter_space.dim
        reversed_dim_sizes = list(reversed(self.parameter_space.dimensional_sizes))
        lower_dims = self.parameter_space.lower_element_num_by_dim()
        reversed_loomed_indices = list(reversed(self.parameter_space.loom_by_flatten_index(flatten_index, lower_dims)))

        total = self.parameter_space.total
        if self.parameter_space.is_infinite():
            msg = "Cannot use this method on infinite space"
            raise LD2InvalidSpaceError(msg)
        reversed_lower_dims = [*list(reversed(lower_dims)), total]

        available_max_upper_reverse_dim = 0  # larger, upper
        for dim, lower_dim in enumerate(lower_dims):
            if flatten_index % lower_dim == 0:
                available_max_upper_reverse_dim = dims - dim - 1
                break

        ticks = [flatten_index + 1]
        for reverse_dim in range(available_max_upper_reverse_dim + 1):
            lower_dim = reversed_lower_dims[reverse_dim]

            size = reversed_dim_sizes[reverse_dim] - reversed_loomed_indices[reverse_dim]
            if size <= 1:
                continue
            d_init = ticks[-1]
            for x in range(1, size):
                tick = d_init + lower_dim * x
                ticks.append(tick)
        return tuple(ticks)

    def _generate_available_next_infinite(self, flatten_index: int) -> tuple[tuple[int, ...], bool]:
        dims = self.parameter_space.dim
        reversed_dim_sizes = list(reversed(self.parameter_space.dimensional_sizes))
        lower_dims = self.parameter_space.lower_element_num_by_dim()
        reversed_loomed_indices = list(reversed(self.parameter_space.loom_by_flatten_index(flatten_index, lower_dims)))

        if not self.parameter_space.is_infinite():
            msg = "Cannot use this method on finite space"
            raise LD2InvalidSpaceError(msg)

        reversed_lower_dims = list(reversed(lower_dims))

        available_max_upper_reverse_dim = 0  # larger, upper
        for dim, lower_dim in enumerate(lower_dims):
            if flatten_index % lower_dim == 0:
                available_max_upper_reverse_dim = dims - dim - 1
                break

        ticks = [flatten_index + 1]
        for reverse_dim in range(available_max_upper_reverse_dim + 1):
            lower_dim = reversed_lower_dims[reverse_dim]
            if reversed_dim_sizes[reverse_dim] is None:
                break

            size = reversed_dim_sizes[reverse_dim] - reversed_loomed_indices[reverse_dim]
            if size <= 1:
                continue
            d_init = ticks[-1]
            for x in range(1, size):
                tick = d_init + lower_dim * x
                ticks.append(tick)

        is_infinitely_available = ticks[-1] - flatten_index == lower_dims[0]
        return tuple(ticks), is_infinitely_available

    @staticmethod
    def _infinite_available_generator(init: tuple[int, ...], ratio: int) -> Generator[int, Any, None]:
        yield from init
        last_v = init[-1]
        for i in itertools.count(1):
            yield last_v + ratio * i

    def to_model(self) -> SuggestStrategyModel:
        return SuggestStrategyModel(
            type="sequential",
            suggest_strategy_param=self.suggest_parameter,
        )



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Annotated, Literal

from pydantic import BaseModel, Field

from lite_dist2.common import float2hex, hex2float, hex2int, int2hex, numerize, portablize
from lite_dist2.expections import LD2InvalidSpaceError, LD2ParameterError
from lite_dist2.interfaces import Mergeable
from lite_dist2.type_definitions import PortableValueType, PrimitiveValueType

if TYPE_CHECKING:
    from collections.abc import Generator


class LineSegmentModel(BaseModel):
    name: str | None = None
    type: Literal["bool", "int", "float"]
    size: str | None
    step: PortableValueType
    start: PortableValueType
    ambient_index: str
    ambient_size: str | None = None
    is_dummy: bool = False


class LineSegment(BaseModel, Mergeable, metaclass=abc.ABCMeta):
    name: str | None = None
    type: Literal["bool", "int", "float"]
    size: int | None
    step: PrimitiveValueType
    start: PrimitiveValueType
    ambient_index: int
    ambient_size: int | None

    @abc.abstractmethod
    def __hash__(self) -> int:
        pass

    @abc.abstractmethod
    def grid(self) -> Generator[PrimitiveValueType, None, None]:
        pass

    @abc.abstractmethod
    def indexed_grid(self) -> Generator[tuple[int, PrimitiveValueType], None, None]:
        pass

    @abc.abstractmethod
    def slice(self, start_index: int, size: int) -> LineSegment:
        pass

    @abc.abstractmethod
    def get_step(self) -> PrimitiveValueType:
        pass

    @abc.abstractmethod
    def merge(self, other: LineSegment, *args: object) -> LineSegment:
        pass

    @abc.abstractmethod
    def to_dummy(self) -> DummyLineSegment:
        pass

    def derived_by_same_ambient_space_with(self, other: LineSegment) -> bool:
        return (
            (self.name == other.name)
            and (self.type == other.type)
            and (self.get_step() == other.get_step())
            and (self.ambient_size == other.ambient_size)
        )

    def can_merge(self, other: LineSegment, *_: object) -> bool:
        if self.ambient_index < other.ambient_index:
            smaller = self
            larger = other
        else:
            smaller = other
            larger = self

        return smaller.end_index() + 1 >= larger.ambient_index

    def end_index(self) -> int:
        if self.size is None:
            msg = "Cannot get end index because this axis is infinite."
            raise LD2InvalidSpaceError(msg)
        return self.ambient_index + self.size - 1

    def is_universal(self) -> bool:
        if self.ambient_size is None:
            return False
        return self.size == self.ambient_size

    @abc.abstractmethod
    def to_model(self) -> LineSegmentModel:
        pass

    @staticmethod
    @abc.abstractmethod
    def from_model(line_segment_model: LineSegmentModel) -> LineSegment:
        pass


class DummyLineSegment(LineSegment):
    name: str
    type: Literal["bool", "int", "float"]
    step: PrimitiveValueType
    start: Literal[0] = 0
    size: Literal[1] = 1
    ambient_index: Literal[0] = 0

    def __hash__(self) -> int:
        return hash((self.name, self.type, self.size, self.ambient_size, self.ambient_index))

    def grid(self, _start_index: int = 0) -> Generator[PrimitiveValueType, None, None]:
        yield from ()

    def indexed_grid(self, _start_index: int = 0) -> Generator[tuple[int, PrimitiveValueType], None, None]:
        yield from ()

    def slice(self, _start_index: int, _size: int) -> DummyLineSegment:
        return self

    def get_step(self) -> PrimitiveValueType:
        return 1

    def get_start_index(self, *_: object) -> int:
        return 0

    def can_merge(self, _other: LineSegment, *_: object) -> bool:
        return False

    def merge(self, _other: LineSegment, *_: object) -> LineSegment:
        return self

    def to_dummy(self) -> DummyLineSegment:
        return self

    def to_model(self) -> LineSegmentModel:
        return LineSegmentModel(
            name=self.name,
            type=self.type,
            size=int2hex(self.size),
            step=portablize(self.type, self.get_step()),
            start="0x0",
            ambient_index="0x0",
            ambient_size=int2hex(self.ambient_size),
            is_dummy=True,
        )

    @staticmethod
    def from_model(line_segment_model: LineSegmentModel) -> DummyLineSegment:
        return DummyLineSegment(
            name=line_segment_model.name,
            type=line_segment_model.type,
            step=numerize(line_segment_model.type, line_segment_model.step),
            ambient_size=None if line_segment_model.ambient_size is None else hex2int(line_segment_model.ambient_size),
        )


class ParameterRangeBool(LineSegment):
    type: Literal["bool"]
    start: bool
    size: Annotated[int, Field(..., ge=1, le=2)]
    step: Annotated[int, Field(1, ge=1, le=1)]
    step: int = 1
    ambient_size: Annotated[int, Field(..., ge=1, le=2)]

    def __hash__(self) -> int:
        return hash((self.name, self.type, self.size, self.start, self.step, self.ambient_size, self.ambient_index))

    def grid(self) -> Generator[PrimitiveValueType, None, None]:
        for i in range(self.size):
            yield bool(int(self.start) + i)

    def indexed_grid(self) -> Generator[tuple[int, PrimitiveValueType], None, None]:
        for i in range(self.size):
            yield i + self.ambient_index, bool(int(self.start) + i)

    def slice(self, start_index: int, size: int) -> ParameterRangeBool:
        if size > self.size - start_index:
            msg = f"{size=}"
            raise LD2ParameterError(msg, "larger than ambient")
        return ParameterRangeBool(
            name=self.name,
            type="bool",
            start=bool(self.start + start_index * self.step),
            size=size,
            step=self.step,
            ambient_index=self.ambient_index + start_index,
            ambient_size=self.ambient_size,
        )

    def get_step(self) -> PrimitiveValueType:
        return self.step

    def get_start_index(self, *_: object) -> int:
        return self.ambient_index

    def merge(self, other: ParameterRangeBool, *_: object) -> ParameterRangeBool:
        smaller, larger = (self, other) if self.ambient_index < other.ambient_index else (other, self)
        size = larger.end_index() - smaller.ambient_index + 1
        return ParameterRangeBool(
            name=self.name,
            type="bool",
            size=size,
            ambient_index=smaller.ambient_index,
            ambient_size=self.ambient_size,
            start=smaller.start,
            step=self.step,
        )

    def to_dummy(self) -> DummyLineSegment:
        return DummyLineSegment(
            name=self.name,
            type=self.type,
            step=self.step,
            ambient_size=self.ambient_size,
        )

    def to_model(self) -> LineSegmentModel:
        return LineSegmentModel(
            name=self.name,
            type=self.type,
            start=self.start,
            size=int2hex(self.size),
            step=int2hex(self.step),
            ambient_index=int2hex(self.ambient_index),
            ambient_size=int2hex(self.ambient_size),
        )

    @staticmethod
    def from_model(line_segment_model: LineSegmentModel) -> ParameterRangeBool:
        return ParameterRangeBool(
            name=line_segment_model.name,
            type="bool",
            size=hex2int(line_segment_model.size),
            ambient_index=hex2int(line_segment_model.ambient_index),
            ambient_size=None if line_segment_model.ambient_size is None else hex2int(line_segment_model.ambient_size),
            start=line_segment_model.start,
            step=hex2int(line_segment_model.step),
        )


class ParameterRangeInt(LineSegment):
    type: Literal["int"]
    start: int
    size: int | None
    step: Annotated[int, Field(1, ge=1)]
    step: int = 1

    def __hash__(self) -> int:
        return hash((self.name, self.type, self.size, self.start, self.step, self.ambient_size, self.ambient_index))

    def grid(self) -> Generator[PrimitiveValueType, None, None]:
        i = 0
        while self.size is None or i < self.size:
            yield self.start + i * self.step
            i += 1

    def indexed_grid(self) -> Generator[tuple[int, PrimitiveValueType], None, None]:
        i = 0
        while self.size is None or i < self.size:
            yield i + self.ambient_index, self.start + i * self.step
            i += 1

    def slice(self, start_index: int, size: int) -> ParameterRangeInt:
        if (self.size is not None) and (size > self.size - start_index):
            msg = f"{size=}"
            raise LD2ParameterError(msg, "larger than ambient")
        return ParameterRangeInt(
            name=self.name,
            type="int",
            start=self.start + start_index * self.step,
            size=size,
            step=self.step,
            ambient_index=self.ambient_index + start_index,
            ambient_size=self.ambient_size,
        )

    def get_step(self) -> PrimitiveValueType:
        return self.step

    def get_start_index(self, *_: object) -> int:
        return self.ambient_index

    def merge(self, other: ParameterRangeInt, *_: object) -> ParameterRangeInt:
        smaller, larger = (self, other) if self.ambient_index < other.ambient_index else (other, self)
        size = larger.end_index() - smaller.ambient_index + 1
        return ParameterRangeInt(
            name=self.name,
            type="int",
            size=size,
            ambient_index=smaller.ambient_index,
            ambient_size=self.ambient_size,
            start=smaller.start,
            step=self.step,
        )

    def to_dummy(self) -> DummyLineSegment:
        return DummyLineSegment(
            name=self.name,
            type=self.type,
            step=self.step,
            ambient_size=self.ambient_size,
        )

    def to_model(self) -> LineSegmentModel:
        return LineSegmentModel(
            name=self.name,
            type=self.type,
            start=int2hex(self.start),
            size=None if self.size is None else int2hex(self.size),
            step=int2hex(self.step),
            ambient_index=int2hex(self.ambient_index),
            ambient_size=None if self.ambient_size is None else int2hex(self.ambient_size),
        )

    @staticmethod
    def from_model(line_segment_model: LineSegmentModel) -> ParameterRangeInt:
        return ParameterRangeInt(
            name=line_segment_model.name,
            type="int",
            size=None if line_segment_model.size is None else hex2int(line_segment_model.size),
            ambient_index=hex2int(line_segment_model.ambient_index),
            ambient_size=None if line_segment_model.ambient_size is None else hex2int(line_segment_model.ambient_size),
            start=hex2int(line_segment_model.start),
            step=hex2int(line_segment_model.step),
        )


class ParameterRangeFloat(LineSegment):
    type: Literal["float"]
    start: float
    size: int | None
    step: Annotated[float, Field(..., gt=0)]

    def __hash__(self) -> int:
        return hash((self.name, self.type, self.size, self.start, self.step, self.ambient_size, self.ambient_index))

    def grid(self) -> Generator[PrimitiveValueType, None, None]:
        i = 0
        ambient_start = self.start - self.ambient_index * self.step
        while self.size is None or i < self.size:
            yield ambient_start + (i + self.ambient_index) * self.step
            i += 1

    def indexed_grid(self) -> Generator[tuple[int, PrimitiveValueType], None, None]:
        i = 0
        ambient_start = self.start - self.ambient_index * self.step
        while self.size is None or i < self.size:
            ambient_index = i + self.ambient_index
            yield ambient_index, ambient_start + ambient_index * self.step
            i += 1

    def slice(self, start_index: int, size: int) -> ParameterRangeFloat:
        if size > self.size - start_index:
            msg = f"{size=}"
            raise LD2ParameterError(msg, "larger than ambient")
        return ParameterRangeFloat(
            name=self.name,
            type="float",
            start=self.start + start_index * self.step,
            size=size,
            step=self.step,
            ambient_index=self.ambient_index + start_index,
            ambient_size=self.ambient_size,
        )

    def get_step(self) -> PrimitiveValueType:
        return self.step

    def get_start_index(self, *_: object) -> int:
        return self.ambient_index

    def merge(self, other: LineSegment, *_: object) -> LineSegment:
        smaller, larger = (self, other) if self.ambient_index < other.ambient_index else (other, self)
        size = larger.end_index() - smaller.ambient_index + 1
        return ParameterRangeFloat(
            name=self.name,
            type="float",
            size=size,
            ambient_index=smaller.ambient_index,
            ambient_size=self.ambient_size,
            start=smaller.start,
            step=self.step,
        )

    def to_dummy(self) -> DummyLineSegment:
        return DummyLineSegment(
            name=self.name,
            type=self.type,
            step=self.step,
            ambient_size=self.ambient_size,
        )

    def to_model(self) -> LineSegmentModel:
        return LineSegmentModel(
            name=self.name,
            type=self.type,
            start=float2hex(self.start),
            size=None if self.size is None else int2hex(self.size),
            step=float2hex(self.step),
            ambient_index=int2hex(self.ambient_index),
            ambient_size=None if self.ambient_size is None else int2hex(self.ambient_size),
        )

    @staticmethod
    def from_model(line_segment_model: LineSegmentModel) -> ParameterRangeFloat:
        return ParameterRangeFloat(
            name=line_segment_model.name,
            type="float",
            size=None if line_segment_model.size is None else hex2int(line_segment_model.size),
            ambient_index=hex2int(line_segment_model.ambient_index),
            ambient_size=None if line_segment_model.ambient_size is None else hex2int(line_segment_model.ambient_size),
            start=hex2float(line_segment_model.start),
            step=hex2float(line_segment_model.step),
        )



from __future__ import annotations

import abc


class Mergeable(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_start_index(self, *args: object) -> int:
        pass

    @abc.abstractmethod
    def can_merge(self, other: Mergeable, *args: object) -> bool:
        pass

    @abc.abstractmethod
    def merge(self, other: Mergeable, *args: object) -> Mergeable:
        pass



import pytest

from lite_dist2.expections import LD2UndefinedError
from lite_dist2.type_definitions import PortableValueType, PrimitiveValueType
from lite_dist2.value_models.base_space import FlattenSegment
from lite_dist2.value_models.jagged_space import ParameterJaggedSpace, ParameterJaggedSpaceModel
from lite_dist2.value_models.line_segment import DummyLineSegment, LineSegmentModel


def test_parameter_jagged_space_hash() -> None:
    space = ParameterJaggedSpace(
        parameters=[(1,)],
        ambient_indices=[(1,)],
        axes_info=[
            DummyLineSegment(name="x", type="int", step=1, ambient_size=100),
        ],
    )
    _ = hash(space)


@pytest.mark.parametrize(
    ("space", "expected"),
    [
        pytest.param(
            ParameterJaggedSpace(
                parameters=[(1,)],
                ambient_indices=[(1,)],
                axes_info=[
                    DummyLineSegment(name="x", type="int", step=1, ambient_size=100),
                ],
            ),
            (1,),
            id="1D",
        ),
        pytest.param(
            ParameterJaggedSpace(
                parameters=[(1,), (2,)],
                ambient_indices=[(1,), (2,)],
                axes_info=[
                    DummyLineSegment(name="x", type="int", step=1, ambient_size=100),
                ],
            ),
            (1,),
            id="1D(parameter affect nothing)",
        ),
        pytest.param(
            ParameterJaggedSpace(
                parameters=[(False, 0, 0)],
                ambient_indices=[(0, 0, 0)],
                axes_info=[
                    DummyLineSegment(name="tf", type="bool", step=1, ambient_size=2),
                    DummyLineSegment(name="x", type="int", step=1, ambient_size=100),
                    DummyLineSegment(name="y", type="int", step=1, ambient_size=70),
                ],
            ),
            (7000, 70, 1),
            id="Multi D",
        ),
    ],
)
def test_parameter_jagged_space_lower_element_num_by_dim(
    space: ParameterJaggedSpace,
    expected: tuple[int, ...],
) -> None:
    actual = space.lower_element_num_by_dim()
    assert actual == expected


@pytest.mark.parametrize(
    ("space", "expected"),
    [
        pytest.param(
            ParameterJaggedSpace(
                parameters=[(0,), (1,)],
                ambient_indices=[(0,), (1,)],
                axes_info=[
                    DummyLineSegment(name="x", type="int", step=1, ambient_size=100),
                ],
            ),
            [
                FlattenSegment(start=0, size=1),
                FlattenSegment(start=1, size=1),
            ],
            id="1D",
        ),
        pytest.param(
            ParameterJaggedSpace(
                parameters=[(78, 1), (1, 78)],
                ambient_indices=[(78, 1), (1, 78)],
                axes_info=[
                    DummyLineSegment(name="x", type="int", step=1, ambient_size=100),
                    DummyLineSegment(name="y", type="int", step=1, ambient_size=100),
                ],
            ),
            [
                FlattenSegment(start=7801, size=1),
                FlattenSegment(start=178, size=1),
            ],
            id="2D",
        ),
    ],
)
def test_parameter_jagged_space_get_flatten_ambient_start_and_size_list(
    space: ParameterJaggedSpace,
    expected: list[FlattenSegment],
) -> None:
    actual = space.get_flatten_ambient_start_and_size_list()
    assert actual == expected


@pytest.mark.parametrize(
    "model",
    [
        pytest.param(
            ParameterJaggedSpaceModel(
                type="jagged",
                parameters=[("0x4e", "0x1"), ("0x1", "0x4e")],
                ambient_indices=[("0x4e", "0x1"), ("0x1", "0x4e")],
                axes_info=[
                    LineSegmentModel(
                        name="x",
                        type="int",
                        size="0x1",
                        step="0x1",
                        start="0x0",
                        ambient_index="0x0",
                        ambient_size="0x64",
                        is_dummy=True,
                    ),
                ],
            ),
        ),
    ],
)
def test_parameter_jagged_space_to_model_from_model(model: ParameterJaggedSpaceModel) -> None:
    space = ParameterJaggedSpace.from_model(model)
    reconstructed = space.to_model()
    assert model == reconstructed


@pytest.mark.parametrize(
    ("primitive", "expected"),
    [
        pytest.param(False, False, id="bool"),
        pytest.param(100, "0x64", id="int"),
        pytest.param(0.25, "0x1.0000000000000p-2", id="float"),
    ],
)
def test_parameter_jagged_space_primitive_to_portable(
    primitive: PrimitiveValueType,
    expected: PortableValueType,
) -> None:
    actual = ParameterJaggedSpace._primitive_to_portable(primitive)
    assert actual == expected


def test_parameter_jagged_space_primitive_to_portable_raise_undefined_type() -> None:
    with pytest.raises(LD2UndefinedError):
        # noinspection PyTypeChecker
        ParameterJaggedSpace._primitive_to_portable([])


@pytest.mark.parametrize(
    ("portable", "expected"),
    [
        pytest.param(False, False, id="bool"),
        pytest.param("0x64", 100, id="int"),
        pytest.param("0x1.0000000000000p-2", 0.25, id="float"),
    ],
)
def test_parameter_jagged_space_portable_to_primitive(
    portable: PortableValueType,
    expected: PrimitiveValueType,
) -> None:
    actual = ParameterJaggedSpace._portable_to_primitive(portable)
    assert actual == expected


def test_parameter_jagged_space_portable_to_primitive_raise_undefined_type() -> None:
    with pytest.raises(LD2UndefinedError):
        # noinspection PyTypeChecker
        ParameterJaggedSpace._portable_to_primitive([])



import threading

import pytest

from lite_dist2.curriculum_models.study import Study
from lite_dist2.curriculum_models.study_portables import StudyModel
from lite_dist2.curriculum_models.study_status import StudyStatus
from lite_dist2.curriculum_models.trial import Mapping, TrialModel, TrialStatus
from lite_dist2.curriculum_models.trial_table import TrialTable, TrialTableModel
from lite_dist2.study_strategies import StudyStrategyModel
from lite_dist2.study_strategies.all_calculation_study_strategy import AllCalculationStudyStrategy
from lite_dist2.suggest_strategies import SequentialSuggestStrategy
from lite_dist2.suggest_strategies.base_suggest_strategy import SuggestStrategyModel, SuggestStrategyParam
from lite_dist2.value_models.aligned_space import ParameterAlignedSpace, ParameterAlignedSpaceModel
from lite_dist2.value_models.line_segment import LineSegmentModel, ParameterRangeFloat
from lite_dist2.value_models.point import ScalarValue
from tests.const import DT


@pytest.mark.parametrize(
    "model",
    [
        pytest.param(
            StudyModel(
                study_id="01",
                name="s1",
                required_capacity=set(),
                status=StudyStatus.running,
                registered_timestamp=DT,
                study_strategy=StudyStrategyModel(type="all_calculation", study_strategy_param=None),
                suggest_strategy=SuggestStrategyModel(
                    type="sequential",
                    suggest_strategy_param=SuggestStrategyParam(strict_aligned=True),
                ),
                parameter_space=ParameterAlignedSpaceModel(
                    type="aligned",
                    axes=[
                        LineSegmentModel(
                            name="x",
                            type="int",
                            size="0x64",
                            step="0x1",
                            start="0x0",
                            ambient_size="0x64",
                            ambient_index="0x0",
                        ),
                        LineSegmentModel(
                            name="y",
                            type="int",
                            size="0x64",
                            step="0x1",
                            start="0x0",
                            ambient_size="0x64",
                            ambient_index="0x0",
                        ),
                    ],
                    check_lower_filling=True,
                ),
                result_type="scalar",
                result_value_type="int",
                trial_table=TrialTableModel(
                    trials=[
                        TrialModel(
                            study_id="01",
                            trial_id="01",
                            timestamp=DT,
                            trial_status=TrialStatus.done,
                            parameter_space=ParameterAlignedSpaceModel(
                                type="aligned",
                                axes=[
                                    LineSegmentModel(
                                        name="x",
                                        type="int",
                                        size="0x1",
                                        step="0x1",
                                        start="0x0",
                                        ambient_size="0x64",
                                        ambient_index="0x0",
                                    ),
                                    LineSegmentModel(
                                        name="y",
                                        type="int",
                                        size="0x2",
                                        step="0x1",
                                        start="0x0",
                                        ambient_size="0x64",
                                        ambient_index="0x0",
                                    ),
                                ],
                                check_lower_filling=True,
                            ),
                            result_type="scalar",
                            result_value_type="float",
                            worker_node_name="w01",
                            worker_node_id="w01",
                            results=[
                                Mapping(
                                    params=(
                                        ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                        ScalarValue(type="scalar", value_type="int", value="0x0", name="y"),
                                    ),
                                    result=ScalarValue(
                                        type="scalar",
                                        value_type="float",
                                        value="0x1.0000000000000p-1",
                                    ),
                                ),
                                Mapping(
                                    params=(
                                        ScalarValue(type="scalar", value_type="int", value="0x0", name="x"),
                                        ScalarValue(type="scalar", value_type="int", value="0x1", name="y"),
                                    ),
                                    result=ScalarValue(
                                        type="scalar",
                                        value_type="float",
                                        value="0x1.0000000000000p-2",
                                    ),
                                ),
                            ],
                        ),
                        TrialModel(
                            study_id="01",
                            trial_id="01",
                            timestamp=DT,
                            trial_status=TrialStatus.running,
                            parameter_space=ParameterAlignedSpaceModel(
                                type="aligned",
                                axes=[
                                    LineSegmentModel(
                                        name="x",
                                        type="int",
                                        size="0x1",
                                        step="0x1",
                                        start="0x0",
                                        ambient_size="0x64",
                                        ambient_index="0x0",
                                    ),
                                    LineSegmentModel(
                                        name="y",
                                        type="int",
                                        size="0x2",
                                        step="0x1",
                                        start="0x2",
                                        ambient_size="0x64",
                                        ambient_index="0x2",
                                    ),
                                ],
                                check_lower_filling=True,
                            ),
                            result_type="scalar",
                            result_value_type="float",
                            worker_node_name="w01",
                            worker_node_id="w01",
                        ),
                    ],
                    aggregated_parameter_space={
                        -1: [],
                        0: [
                            ParameterAlignedSpaceModel(
                                type="aligned",
                                axes=[
                                    LineSegmentModel(
                                        name="x",
                                        type="int",
                                        size="0x1",
                                        step="0x1",
                                        start="0x0",
                                        ambient_size="0x64",
                                        ambient_index="0x0",
                                    ),
                                    LineSegmentModel(
                                        name="y",
                                        type="int",
                                        size="0x2",
                                        step="0x1",
                                        start="0x0",
                                        ambient_size="0x64",
                                        ambient_index="0x0",
                                    ),
                                ],
                                check_lower_filling=True,
                            ),
                        ],
                    },
                ),
            ),
            id="full_definition",
        ),
    ],
)
def test_study_to_model_from_model(model: StudyModel) -> None:
    study = Study.from_model(model)
    reconstructed = study.to_model()
    assert model == reconstructed


def test_study_suggest_receipt_single_thread() -> None:
    _parameter_space = ParameterAlignedSpace(
        axes=[
            ParameterRangeFloat(
                name="x",
                type="float",
                size=20,
                step=0.5,
                start=0.0,
                ambient_index=0,
                ambient_size=20,
            ),
            ParameterRangeFloat(
                name="y",
                type="float",
                size=20,
                step=0.5,
                start=0.0,
                ambient_index=0,
                ambient_size=20,
            ),
        ],
        check_lower_filling=True,
    )
    study = Study(
        study_id="s01",
        name="synchronous_test",
        required_capacity=set(),
        status=StudyStatus.running,
        registered_timestamp=DT,
        study_strategy=AllCalculationStudyStrategy(study_strategy_param=None),
        suggest_strategy=SequentialSuggestStrategy(
            suggest_parameter=SuggestStrategyParam(strict_aligned=True),
            parameter_space=_parameter_space,
        ),
        parameter_space=_parameter_space,
        result_type="scalar",
        result_value_type="float",
        trial_table=TrialTable(trials=[], aggregated_parameter_space=None),
    )

    def contract_and_submit() -> None:
        # trial 取得
        trial = study.suggest_next_trial(num=5, worker_node_name="w01", worker_node_id="w01")
        if trial is None:
            return

        # 結果書き込み
        raw_mappings = []
        for parameter in trial.parameter_space.grid():
            dummy_result = 0.5
            raw_mappings.append((parameter, dummy_result))
        trial.set_result(trial.convert_mappings_from(raw_mappings))

        # trial 送信
        study.receipt_trial(trial)

    while not study.is_done():
        contract_and_submit()

    expected_trial_num = 80  # 20*20/5
    assert study.trial_table.count_trial() == expected_trial_num


def test_study_suggest_receipt_multi_threads_synchronous() -> None:
    _parameter_space = ParameterAlignedSpace(
        axes=[
            ParameterRangeFloat(
                name="x",
                type="float",
                size=20,
                step=0.5,
                start=0.0,
                ambient_index=0,
                ambient_size=20,
            ),
            ParameterRangeFloat(
                name="y",
                type="float",
                size=20,
                step=0.5,
                start=0.0,
                ambient_index=0,
                ambient_size=20,
            ),
        ],
        check_lower_filling=True,
    )
    study = Study(
        study_id="s01",
        name="synchronous_test",
        required_capacity=set(),
        status=StudyStatus.running,
        registered_timestamp=DT,
        study_strategy=AllCalculationStudyStrategy(study_strategy_param=None),
        suggest_strategy=SequentialSuggestStrategy(
            suggest_parameter=SuggestStrategyParam(strict_aligned=True),
            parameter_space=_parameter_space,
        ),
        parameter_space=_parameter_space,
        result_type="scalar",
        result_value_type="float",
        trial_table=TrialTable(trials=[], aggregated_parameter_space=None),
    )

    def contract_and_submit() -> None:
        # trial 取得
        trial = study.suggest_next_trial(num=5, worker_node_name="w01", worker_node_id="w01")

        # 結果書き込み
        raw_mappings = []
        for parameter in trial.parameter_space.grid():
            dummy_result = 0.5
            raw_mappings.append((parameter, dummy_result))
        trial.set_result(trial.convert_mappings_from(raw_mappings))

        # trial 送信
        study.receipt_trial(trial)

    # スレッドを複数作成

    num_threads = 10
    while not study.is_done():
        threads = []
        for i in range(num_threads):
            thread = threading.Thread(target=contract_and_submit, name=f"Thread-{i + 1}")
            threads.append(thread)

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

    expected_trial_num = 80  # 20*20/5
    assert study.trial_table.count_trial() == expected_trial_num



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import requests

from lite_dist2.curriculum_models.study_status import StudyStatus
from lite_dist2.curriculum_models.trial import Trial
from lite_dist2.expections import LD2TableNodeClientError, LD2TableNodeServerError
from lite_dist2.table_node_api.table_param import StudyRegisterParam, TrialRegisterParam, TrialReserveParam
from lite_dist2.table_node_api.table_response import (
    OkResponse,
    StudyRegisteredResponse,
    StudyResponse,
    TrialReserveResponse,
)

if TYPE_CHECKING:
    from typing import Any, ClassVar

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class TableNodeClient:
    INSTANT_API_TIMEOUT_SECONDS = 10
    HEADERS: ClassVar[dict[str, str]] = {"Content-Type": "application/json; charset=utf-8"}

    def __init__(self, ip: str, port: int | str) -> None:
        self.domain = f"http://{ip}:{port}"

    def ping(self) -> bool:
        try:
            _ = self._get("/ping", timeout=self.INSTANT_API_TIMEOUT_SECONDS)
        except LD2TableNodeServerError:
            return False
        return True

    def register_study(self, param: StudyRegisterParam) -> StudyRegisteredResponse:
        _, d = self._post("/study/register", self.INSTANT_API_TIMEOUT_SECONDS, param.model_dump(mode="json"))

        resp = StudyRegisteredResponse.model_validate(d)
        logger.info("Registered study: %s", resp.study_id)
        return resp

    def reserve_trial(
        self,
        worker_id: str,
        worker_name: str | None,
        max_size: int,
        retaining_capacity: set[str],
        timeout_seconds: int,
    ) -> Trial | None:
        param = TrialReserveParam(
            retaining_capacity=retaining_capacity,
            max_size=max_size,
            worker_node_name=worker_name,
            worker_node_id=worker_id,
        )
        status_code, d = self._post("/trial/reserve", timeout_seconds, param.model_dump(mode="json"))

        resp = TrialReserveResponse.model_validate(d)
        if status_code == requests.codes.accepted or resp.trial is None:
            logger.info("Cannot reserve trial")
            return None

        trial = Trial.from_model(resp.trial)
        logger.info("Reserved trial (size=%d)", trial.parameter_space.get_total())
        return trial

    def register_trial(self, trial: Trial, timeout_seconds: int) -> None:
        param = TrialRegisterParam(trial=trial.to_model())
        status_code, _ = self._post("/trial/register", timeout_seconds, param.model_dump(mode="json"))
        if status_code != requests.codes.ok:
            logger.warning("Failed to register trial. This trial might be timed out or study might be cancelled.")

    def study(self, study_id: str | None = None, name: str | None = None) -> StudyResponse | None:
        _, resp = self._get("/study", self.INSTANT_API_TIMEOUT_SECONDS, {"study_id": study_id, "name": name})
        study_response = StudyResponse.model_validate(resp)
        if study_response.status != StudyStatus.done:
            detail_info = f"{study_id=}" if study_id is not None else f"{name=}"
            logger.info("Study(%s) is %s", detail_info, str(study_response))
        return study_response

    def save(self) -> OkResponse:
        _, resp = self._get("/save", self.INSTANT_API_TIMEOUT_SECONDS)
        return OkResponse.model_validate(resp)

    def _get(self, path: str, timeout: int, query: dict[str, str] | None = None) -> tuple[int, dict[str, Any]]:
        url = f"{self.domain}{path}"
        response = requests.get(
            url,
            headers=self.HEADERS,
            params=query,
            timeout=timeout,
        )
        return response.status_code, self._check_status_code(response)

    def _post(self, path: str, timeout_seconds: int, body: dict[str, Any]) -> tuple[int, dict[str, Any]]:
        url = f"{self.domain}{path}"
        response = requests.post(
            url,
            headers=self.HEADERS,
            json=body,
            timeout=timeout_seconds,
        )
        return response.status_code, self._check_status_code(response)

    @staticmethod
    def _check_status_code(response: requests.Response) -> dict[str, Any]:
        if response.status_code >= requests.codes.internal_server_error:
            raise LD2TableNodeServerError(response.text)
        if response.status_code >= requests.codes.bad_request:
            raise LD2TableNodeClientError(response.text)
        return response.json()
